from __future__ import annotations

from logging import getLogger
from typing import Any, Callable, Tuple

import torch
from torch import Tensor, no_grad
from torch.autograd import Function

from pyqtorch.circuit import QuantumCircuit
from pyqtorch.composite import Scale
from pyqtorch.embed import Embedding
from pyqtorch.hamiltonians import HamiltonianEvolution, Observable
from pyqtorch.matrices import DEFAULT_REAL_DTYPE
from pyqtorch.primitives import Parametric
from pyqtorch.utils import param_dict

logger = getLogger(__name__)


class PSRExpectation(Function):
    r"""
    Implementation of the generalized parameter shift rule.

    Note that only operations with two distinct eigenvalues
    from their generator (i.e., compatible with single_gap_shift)
    are supported at the moment.

    Compared to the original parameter shift rule
    which only works for quantum operations whose generator has a single gap
    in its eigenvalue spectrum, GPSR works with arbitrary
    generators of quantum operations.

    For this, we define the differentiable function as quantum expectation value

    $$
    f(x) = \left\langle 0\right|\hat{U}^{\dagger}(x)\hat{C}\hat{U}(x)\left|0\right\rangle
    $$

    where $\hat{U}(x)={\rm exp}{\left( -i\frac{x}{2}\hat{G}\right)}$
    is the quantum evolution operator with generator $\hat{G}$ representing the structure
    of the underlying quantum circuit and $\hat{C}$ is the cost operator.
    Then using the eigenvalue spectrum $\left\{ \lambda_n\right\}$ of the generator $\hat{G}$
    we calculate the full set of corresponding unique non-zero spectral gaps
    $\left\{ \Delta_s\right\}$ (differences between eigenvalues).
    It can be shown that the final expression of derivative of $f(x)$
    is then given by the following expression:

    $\begin{equation}
    \frac{{\rm d}f\left(x\right)}{{\rm d}x}=\overset{S}{\underset{s=1}{\sum}}\Delta_{s}R_{s},
    \end{equation}$

    where $S$ is the number of unique non-zero spectral gaps and $R_s$ are real quantities that
    are solutions of a system of linear equations

    $\begin{equation}
    \begin{cases}
    F_{1} & =4\overset{S}{\underset{s=1}{\sum}}{\rm sin}
    \left(\frac{\delta_{1}\Delta_{s}}{2}\right)R_{s},\\
    F_{2} & =4\overset{S}{\underset{s=1}{\sum}}{\rm sin}
    \left(\frac{\delta_{2}\Delta_{s}}{2}\right)R_{s},\\
    & ...\\
    F_{S} & =4\overset{S}{\underset{s=1}{\sum}}{\rm sin}
    \left(\frac{\delta_{M}\Delta_{s}}{2}\right)R_{s}.
    \end{cases}
    \end{equation}$

    Here $F_s=f(x+\delta_s)-f(x-\delta_s)$ denotes the difference between values
    of functions evaluated at shifted arguments $x\pm\delta_s$.

    Arguments:
        circuit: A QuantumCircuit instance
        state: A state in the form of [2 * n_qubits + [batch_size]]
        observable: An hermitian operator.
        embedding: An optional instance of `Embedding`.
        expectation_method: Callable for computing expectations.
        param_names: A list of parameter names.
        *param_values: A unpacked tensor of values for each parameter.

    The forward method expects each of the above arguments, computes the expectation value
    and stores all of the above arguments in the 'ctx' (context) object along with the
    the state after applying 'circuit', 'out_state', and the 'projected_state', i.e. after applying
    'observable' to 'out_state'.
    """

    @staticmethod
    @no_grad()
    def forward(
        ctx: Any,
        circuit: QuantumCircuit,
        state: Tensor,
        observable: Observable,
        embedding: Embedding,
        expectation_method: Callable,
        param_names: list[str],
        *param_values: Tensor,
    ) -> Tensor:
        if embedding is not None:
            logger.error("GPSR does not support Embedding.")
        ctx.circuit = circuit
        ctx.observable = observable
        ctx.param_names = param_names
        ctx.state = state
        ctx.embedding = embedding
        values = param_dict(param_names, param_values)
        ctx.save_for_backward(*param_values)
        ctx.expectation_method = expectation_method
        return expectation_method(circuit, state, observable, values, embedding)

    @staticmethod
    def backward(ctx: Any, grad_out: Tensor) -> Tuple[None, ...]:
        """The PSRExpectation backward call.

        Arguments:
            ctx (Any): Context object for accessing stored information.
            grad_out (Tensor): Current jacobian tensor.

        Returns:
            A tuple of updated jacobian tensor.

        Raises:
            ValueError: When operation is not supported.
        """

        values = param_dict(ctx.param_names, ctx.saved_tensors)
        dtype_values = DEFAULT_REAL_DTYPE
        device = torch.device("cpu")
        try:
            dtype_values, device = [(v.dtype, v.device) for v in values.values()][0]
        except Exception:
            pass

        shift_pi2 = torch.tensor(torch.pi, dtype=dtype_values) / 2.0

        def expectation_fn(values: dict[str, Tensor]) -> Tensor:
            """Use the PSRExpectation for nested grad calls.

            Arguments:
                values: Dictionary with parameter values.

            Returns:
                Expectation evaluation.
            """
            return PSRExpectation.apply(
                ctx.circuit,
                ctx.state,
                ctx.observable,
                ctx.embedding,
                ctx.expectation_method,
                values.keys(),
                *values.values(),
            )

        def single_gap_shift(
            param_uuid: str,
            values: dict[str, Tensor],
            spectral_gap: Tensor,
            shift_prefac: float = 1.0,
        ) -> Tensor:
            """Implements single gap PSR rule.

            Args:
                param_uuid (str): Uuid of Parameter to help identify the computed gradient.
                values: Dictionary with parameter values.
                spectral_gap: Spectral gap value for PSR.
                shift_prefac: Shift prefactor value to multiply pi/2.
                Defaults to 1.

            Returns:
                Gradient evaluation for param_name.
            """

            # device conversions
            spectral_gap = spectral_gap.to(device=device)
            shift = shift_pi2.to(device=device) * shift_prefac

            # apply shift rule
            shifted_values = values.copy()
            if param_uuid in shifted_values:
                shifted_values[param_uuid] = shifted_values[param_uuid] + shift
            else:
                shifted_values[param_uuid] = shift
            f_plus = expectation_fn(shifted_values)
            shifted_values[param_uuid] = shifted_values[param_uuid] - 2 * shift
            f_minus = expectation_fn(shifted_values)
            shifted_values[param_uuid] = shifted_values[param_uuid] + shift
            return (
                spectral_gap
                * (f_plus - f_minus)
                / (4.0 * torch.sin(spectral_gap * shift / 2.0))
            )

        def multi_gap_shift(
            param_uuid: str,
            values: dict[str, Tensor],
            spectral_gaps: Tensor,
            shift_prefac: float = 0.5,
        ) -> Tensor:
            """Implement multi gap PSR rule.

            See Kyriienko1 and Elfving, 2021 for details:
            https://arxiv.org/pdf/2108.01218.pdf

            Args:
                param_uuid (str): Uuid of Parameter to help identify the computed gradient.
                values: Dictionary with parameter values.
                spectral_gaps: Spectral gaps value for PSR.
                shift_prefac: Shift prefactor value for PSR shifts.
                Defaults to torch.tensor(0.5).

            Returns:
                Gradient evaluation for param_name.
            """
            n_eqs = len(spectral_gaps)
            dtype = torch.promote_types(dtype_values, spectral_gaps.dtype)
            spectral_gaps = spectral_gaps.to(device=device)
            PI = torch.tensor(torch.pi, dtype=dtype)
            shifts = shift_prefac * torch.linspace(
                PI / 2.0 - PI / 4.0, PI / 2.0 + PI / 5.0, n_eqs, dtype=dtype
            )
            shifts = shifts.to(device=device)

            # calculate F vector and M matrix
            # (see: https://arxiv.org/pdf/2108.01218.pdf on p. 4 for definitions)
            F = []
            M = torch.empty((n_eqs, n_eqs), dtype=dtype).to(device=device)
            batch_size = 1
            shifted_params = values.copy()
            if param_uuid not in values:
                shifted_params[param_uuid] = torch.zeros_like(
                    shifts[0], requires_grad=True
                )
            for i in range(n_eqs):
                # + shift
                shifted_params[param_uuid] = shifted_params[param_uuid] + shifts[i]
                f_plus = expectation_fn(shifted_params)

                # - shift
                shifted_params[param_uuid] = shifted_params[param_uuid] - 2 * shifts[i]
                f_minus = expectation_fn(shifted_params)
                shifted_params[param_uuid] = (
                    shifted_params[param_uuid] + shifts[i]
                )  # reset
                F.append((f_plus - f_minus))

                # calculate M matrix
                for j in range(n_eqs):
                    M[i, j] = 4 * torch.sin(shifts[i] * spectral_gaps[j] / 2)

            # get number of observables from expectation value tensor
            if f_plus.numel() > 1:
                batch_size = F[0].shape[0]

            F = torch.stack(F).reshape(n_eqs, -1)
            R = torch.linalg.solve(M, F)
            dfdx = torch.sum(spectral_gaps * R, dim=0).reshape(batch_size)

            return dfdx

        def vjp(
            param_uuid: str,
            spectral_gap: Tensor,
            values: dict[str, Tensor],
            shift_prefac: float,
        ) -> Tensor:
            """Vector-jacobian product between `grad_out` and jacobians of parameters.

            Args:
                param_uuid (str): Uuid of Parameter to help identify the computed gradient.
                spectral_gap: Spectral gap of the corresponding operation.
                values: Dictionary with parameter values.
                shift_prefac: Shift prefactor value for PSR shifts.

            Returns:
                Updated jacobian by PSR.
            """
            psr_fn = multi_gap_shift if len(spectral_gap) > 1 else single_gap_shift

            return grad_out * psr_fn(  # type: ignore[operator]
                param_uuid,
                values,
                spectral_gap,
                shift_prefac=shift_prefac,
            )

        grads = {p: None for p in values.keys()}

        def update_gradient(
            param_name: str, param_uuid: str, spectral_gap: Tensor, shift_prefac: float
        ):
            """Update gradient of a parameter using PSR.

            Args:
                param_name (str): Parameter name to compute gradient over.
                param_uuid (str): Uuid of Parameter to help identify the computed gradient.
                spectral_gap (Tensor): Spectral gap of the corresponding operation.
                shift_prefac (float): Shift prefactor value for PSR shifts.
            """
            if grads[param_name] is None:
                grads[param_name] = vjp(param_uuid, spectral_gap, values, shift_prefac)
            else:
                grad_contrib = vjp(param_uuid, spectral_gap, values, shift_prefac)
                grads[param_name] += grad_contrib.reshape(grads[param_name].shape)  # type: ignore[attr-defined]

        for op in ctx.circuit.flatten():

            if isinstance(op, (Parametric, HamiltonianEvolution)) and op.is_parametric:
                if values[op.param_name].requires_grad:  # type: ignore[index]
                    factor = 1.0 if isinstance(op, Parametric) else 2.0
                    if len(op.spectral_gap) > 1:
                        update_gradient(
                            op.param_name, op._param_uuid, factor * op.spectral_gap, 0.5  # type: ignore[arg-type]
                        )
                    else:
                        shift_factor = 1.0
                        # note the spectral gap can be empty
                        # this is handled in single-gap PSR
                        if isinstance(op, HamiltonianEvolution):
                            shift_factor = (
                                1.0 / (op.spectral_gap.item() * factor)
                                if len(op.spectral_gap) == 1
                                else 1.0
                            )
                        update_gradient(
                            op.param_name,
                            op._param_uuid,
                            factor * op.spectral_gap,
                            shift_factor,
                        )

        return (
            None,
            None,
            None,
            None,
            None,
            None,
            *[grads[p] for p in ctx.param_names],
        )


def check_support_psr(circuit: QuantumCircuit):
    """Checking that circuit has only compatible operations for PSR.

    Args:
        circuit (QuantumCircuit): Circuit to check.

    Raises:
        ValueError: When circuit contains Scale, HamiltonianEvolution,
                    or a param_name is used multiple times in the circuit.
    """

    param_names = list()
    for op in circuit.flatten():
        if isinstance(op, Scale):
            raise ValueError(
                f"PSR is not applicable as circuit contains an operation of type: {type(op)}."
            )
        if isinstance(op, HamiltonianEvolution) and op.is_parametric_generator:
            raise ValueError(
                f"PSR is not applicable as circuit contains an operation of type: {type(op)} \
                    whose generator type is {op.generator_type}."
            )
        elif isinstance(op, (Parametric, HamiltonianEvolution)) and op.is_parametric:
            param_names.append(op.param_name)
        else:
            continue



from __future__ import annotations

import logging
from collections import Counter, OrderedDict
from dataclasses import dataclass
from enum import Enum
from functools import lru_cache, partial, wraps
from logging import getLogger
from math import sqrt
from string import ascii_uppercase as ABC
from typing import Any, Callable, Sequence

import numpy as np
import torch
from numpy import arange, argsort, array, delete, log2
from numpy import ndarray as NDArray
from torch import Tensor, moveaxis
from typing_extensions import TypeAlias

import pyqtorch as pyq
from pyqtorch.matrices import DEFAULT_MATRIX_DTYPE, DEFAULT_REAL_DTYPE, IDIAG, IMAT

State: TypeAlias = Tensor
Operator: TypeAlias = Tensor

ATOL = 1e-06
ATOL_embedding = 1e-03
RTOL = 0.0
GRADCHECK_ATOL = 1e-04
GRADCHECK_ATOL_hamevo = 1e-03
GRADCHECK_sampling_ATOL = 1e-01
PSR_ACCEPTANCE = 1e-05
ABC_ARRAY: NDArray = array(list(ABC))

logger = getLogger(__name__)


class OrderedCounter(Counter, OrderedDict):  # type: ignore [misc]
    pass


def qubit_support_as_tuple(support: int | tuple[int, ...]) -> tuple[int, ...]:
    """Make sure support returned is a tuple of integers.

    Args:
        support (int | tuple[int, ...]): Qubit support.

    Returns:
        tuple[int, ...]: Qubit support as tuple.
    """
    if isinstance(support, np.integer):
        return (support.item(),)
    qubit_support = (support,) if isinstance(support, int) else tuple(support)
    return qubit_support


def _round_operator(t: Tensor, decimals: int = 4) -> Tensor:
    if torch.is_complex(t):

        def _round(_t: Tensor) -> Tensor:
            r = _t.real.round(decimals=decimals)
            i = _t.imag.round(decimals=decimals)
            return torch.complex(r, i)

    else:

        def _round(_t: Tensor) -> Tensor:
            return _t.round(decimals=decimals)

    fn = torch.vmap(_round)
    return fn(t)


def inner_prod(bra: Tensor, ket: Tensor) -> Tensor:
    """
    Compute the inner product :math:`\\langle\\bra|\\ket\\rangle`

    Arguments:
        bra: left part quantum state.
        ket: right part quantum state.

    Returns:
        The inner product.
    """
    if logger.isEnabledFor(logging.DEBUG):
        logger.debug("Inner prod calculation")
        torch.cuda.nvtx.range_push("inner_prod")

    n_qubits = len(bra.size()) - 1
    bra = bra.reshape((2**n_qubits, bra.size(-1)))
    ket = ket.reshape((2**n_qubits, ket.size(-1)))
    res = torch.einsum("ib,ib->b", bra.conj(), ket)
    if logger.isEnabledFor(logging.DEBUG):
        torch.cuda.nvtx.range_pop()
        logger.debug("Inner prod complete")
    return res


def overlap(bra: Tensor, ket: Tensor) -> Tensor:
    """
    Compute the overlap :math:`|\\langle\\bra|\\ket\\rangle|^2`

    Arguments:
        bra: left part quantum state.
        ket: right part quantum state.

    Returns:
        The overlap.
    """
    return torch.pow(inner_prod(bra, ket).real, 2)


def counts_to_orderedcounter(
    binary_count: Tensor, length_bitstring: int
) -> OrderedCounter:
    """Convert counts (from torch.bincount) to an OrderedCounter.

    Note the output of torch.bincount is ordered
    and having an OrderedCounter can be convenient.

    Args:
        binary_count (Tensor): Counts per bitstring as Tensor.
        length_bitstring (int): Length of the bitstring.

    Returns:
        OrderedCounter: Ordered counter of bitstrings
    """
    return OrderedCounter(
        OrderedDict(
            [
                (format(k, "0{}b".format(length_bitstring)), count.item())
                for k, count in enumerate(binary_count)
                if count > 0
            ]
        )
    )


def sample_multinomial(
    probs: Tensor,
    length_bitstring: int,
    n_samples: int,
    return_counter: bool = True,
    minlength: int = 0,
) -> OrderedCounter | Tensor:
    """Sample bitstrings from a probability distribution.

    Args:
        probs (Tensor): Probability distribution
        length_bitstring (int): Maximal length of bitstring.
        n_samples (int): Number of samples to extract.
        instead of ratios.
        return_counter (bool): If True, return OrderedCounter object.
            Otherwise, the result of torch.bincount is returned.
        minlength (int): minimum number of bins. Should be non-negative.

    Returns:
        OrderedCounter | Tensor: Sampled bitstrings with their frequencies or probabilities.
    """

    bincount_output = torch.bincount(
        torch.multinomial(input=probs, num_samples=n_samples, replacement=True),
        minlength=minlength,
    )

    if return_counter:
        return counts_to_orderedcounter(bincount_output, length_bitstring)
    return bincount_output


class StrEnum(str, Enum):
    def __str__(self) -> str:
        """Used when dumping enum fields in a schema."""
        ret: str = self.value
        return ret

    @classmethod
    def list(cls) -> list[str]:
        return list(map(lambda c: c.value, cls))  # type: ignore


class DiffMode(StrEnum):
    """
    Which Differentiation method to use.
    """

    AD = "ad"
    """Use torch.autograd to perform differentiation."""
    ADJOINT = "adjoint"
    """An implementation of "Efficient calculation of gradients
                                       in classical simulations of variational quantum algorithms",
                                       Jones & Gacon, 2020"""

    GPSR = "gpsr"
    """The generalized parameter-shift rule"""


class DropoutMode(StrEnum):
    """
    Which Dropout mode to use, using the methods stated in https://arxiv.org/abs/2310.04120.

    Options: rotational    - Randomly drops entangling rotational gates.
             entangling    - Randomly drops entangling gates.
             canonical_fwd - Randomly drops rotational gates and next immediate entangling
                            gates whose target bit is located on dropped rotational gates.
    """

    ROTATIONAL = "rotational_dropout"
    ENTANGLING = "entangling_dropout"
    CANONICAL_FWD = "canonical_fwd_dropout"


def is_normalized(state: Tensor, atol: float = ATOL) -> bool:
    """
    Function to test if the probabilities from the state sum up to 1.

    Arguments:
        state: State to test.
        atol: Tolerance for check.

    Returns:
        True if normalized, False otherwise.
    """
    n_qubits = len(state.size()) - 1
    batch_size = state.size()[-1]
    state = state.reshape((2**n_qubits, batch_size))
    sum_probs = (state.abs() ** 2).sum(dim=0)
    ones = torch.ones(batch_size, dtype=DEFAULT_REAL_DTYPE)
    return torch.allclose(sum_probs, ones, rtol=RTOL, atol=atol)  # type: ignore[no-any-return]


def is_diag_batched(H: Operator, atol: Tensor = ATOL, batch_dim: int = -1) -> bool:
    """
    Returns True if the batched tensors of H are diagonal.
    Arguments:
        H: Input tensors.
        atol: Tolerance for near-zero values.
        batch_dim: batch dimension to go over.
    Returns:
        True if diagonal, else False.
    """
    if len(H.shape) > 2:
        diag_check = torch.tensor(
            [is_diag(H[..., i], atol) for i in range(H.shape[batch_dim])],
            device=H.device,
        )
        return bool(torch.prod(diag_check).item())
    else:
        return is_diag(H, atol)


def is_diag(H: Tensor, atol: Tensor = ATOL) -> bool:
    """
    Returns True if tensor H is diagonal.

    Reference: https://stackoverflow.com/questions/43884189/check-if-a-large-matrix-is-diagonal-matrix-in-python

    Arguments:
        H: Input tensor.
        atol: If off-diagonal values are lower than atol in amplitude, H is considered diagonal.

    Returns:
        True if diagonal, else False.
    """
    offdiag_view = H - torch.diag(torch.diag(H))
    return torch.count_nonzero(torch.abs(offdiag_view).gt(atol)).item() == 0


def finitediff(
    f: Callable,
    x: Tensor,
    derivative_indices: tuple[int, ...],
    eps: float | None = None,
) -> Tensor:
    """
    Compute the finite difference of a function at a point.

    Args:
        f: The function to differentiate.
        x: Input of size `(batch_size, input_size)`.
        derivative_indices: Which *input* to differentiate (i.e. which variable x[:,i])
        eps: finite difference spacing (uses `torch.finfo(x.dtype).eps ** (1 / (2 + order))`
            as default)

    Returns:
        (Tensor): The finite difference of the function at the point `x`.
    """

    if eps is None:
        order = len(derivative_indices)
        eps = torch.finfo(x.dtype).eps ** (1 / (2 + order))

    # compute derivative direction vector(s)
    delta = torch.zeros_like(x)
    i = derivative_indices[0]
    delta[:, i] += torch.as_tensor(eps, dtype=x.dtype)
    denominator = 1 / delta[:, i]

    # recursive finite differencing for higher order than 3 / mixed derivatives
    if len(derivative_indices) > 3 or len(set(derivative_indices)) > 1:
        di = derivative_indices[1:]
        return (
            (finitediff(f, x + delta, di) - finitediff(f, x - delta, di))
            * denominator
            / 2
        )
    if len(derivative_indices) == 3:
        return (
            (f(x + 2 * delta) - 2 * f(x + delta) + 2 * f(x - delta) - f(x - 2 * delta))
            * denominator**3
            / 2
        )
    if len(derivative_indices) == 2:
        return (f(x + delta) + f(x - delta) - 2 * f(x)) * denominator**2
    if len(derivative_indices) == 1:
        return (f(x + delta) - f(x - delta)) * denominator / 2
    raise ValueError(
        "If you see this error there is a bug in the `finitediff` function."
    )


def product_state(
    bitstring: str,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> Tensor:
    """
    Create a batch of quantum states :math:`|\\b\\rangle` where b is the bitstring input.

    Arguments:
        bitstring: The bitstring b to represent.
        batch_size: The size of the batch.
        device: Device where tensors are stored.
        dtype: Type of tensors.

    Returns:
        Batch of quantum states representing bitstring.
    """
    state = torch.zeros((2 ** len(bitstring), batch_size), dtype=dtype)
    state[int(bitstring, 2)] = torch.tensor(1.0 + 0j, dtype=dtype)
    return state.reshape([2] * len(bitstring) + [batch_size]).to(device=device)


def zero_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> Tensor:
    """
    Create a batch of :math:`|\\0\\rangle^{\\otimes n}` states defined on n qubits.

    Args:
        n_qubits (int): Number of qubits n the state is defined on.
        batch_size (int, optional): The size of the batch.
            Defaults to 1.
        device (str | torch.device, optional): Device where tensors are stored.
            Defaults to "cpu".
        dtype (torch.dtype, optional): Type of tensors. Defaults to DEFAULT_MATRIX_DTYPE.

    Returns:
        Tensor: Batch of uniform quantum states.
    """
    return product_state("0" * n_qubits, batch_size, dtype=dtype, device=device)


def uniform_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
) -> Tensor:
    """
    Create a batch of uniform states with equal probabilities for computational basis.

    Arguments:
        n_qubits: Number of qubits the state is defined on.
        batch_size: The size of the batch.
        device: Device where tensors are stored.
        dtype: Type of tensors.

    Returns:
        Batch of uniform quantum states.
    """
    state = torch.ones((2**n_qubits, batch_size), dtype=dtype, device=device)
    state = state / torch.sqrt(torch.tensor(2**n_qubits))
    return state.reshape([2] * n_qubits + [batch_size])


def random_state(
    n_qubits: int,
    batch_size: int = 1,
    device: str | torch.device = "cpu",
    dtype: torch.dtype = DEFAULT_MATRIX_DTYPE,
    generator: torch.Generator = None,
) -> Tensor:
    """
    Create a batch of random quantum state.

    Arguments:
        n_qubits: Number of qubits the state is defined on.
        batch_size: The size of the batch.
        device: Device where tensors are stored.
        dtype: Type of tensors.

    Returns:
        Batch of random quantum states.
    """

    def _normalize(wf: Tensor) -> Tensor:
        return wf / torch.sqrt((wf.abs() ** 2).sum())

    def _rand(n_qubits: int) -> Tensor:
        N = 2**n_qubits
        x = -torch.log(torch.rand(N, generator=generator, dtype=dtype))
        sumx = torch.sum(x)
        phases = torch.rand(N, generator=generator, dtype=dtype) * 2.0 * torch.pi
        return _normalize(
            (torch.sqrt(x / sumx) * torch.exp(1j * phases))
            .reshape(N, 1)
            .type(dtype)
            .to(device)
        )

    state = torch.concat(tuple(_rand(n_qubits) for _ in range(batch_size)), dim=1)
    return state.reshape([2] * n_qubits + [batch_size]).to(device=device)


def param_dict(keys: Sequence[str], values: Sequence[Tensor]) -> dict[str, Tensor]:
    """
    Create a dictionary mapping parameters with their values.

    Arguments:
        keys: Parameter names as keys.
        values: Values of parameters.

    Returns:
        Dictionary mapping parameters and values.
    """
    return {key: val for key, val in zip(keys, values)}


class DensityMatrix(Tensor):
    pass


def density_mat(state: Tensor) -> DensityMatrix:
    """
    Computes the density matrix from a pure state vector.

    Arguments:
        state: The pure state vector :math:`|\\psi\\rangle`.

    Returns:
        Tensor: The density matrix :math:`\\rho = |\\psi \\rangle \\langle\\psi|`.
    """
    if isinstance(state, DensityMatrix):
        return state
    n_qubits = len(state.size()) - 1
    batch_size = state.shape[-1]
    state = state.reshape(2**n_qubits, batch_size)
    return DensityMatrix(torch.einsum("ib,jb->ijb", (state, state.conj())))


def operator_kron(op1: Tensor, op2: Tensor) -> Tensor:
    """
    Compute the Kronecker product of two operators.

    Prevents errors related to the shape of the operators
    [2**n_qubits, 2**n_qubits, batch_size] when simply using `torch.kron()`.
    Use of `.contiguous()` to avoid errors related to the `torch.kron()` of a transposing tensor

    Arguments:
        op1: The first input tensor.
        op2: The second input tensor.

    Returns:
        The resulting tensor after applying the Kronecker product
    """
    batch_size_1, batch_size_2 = op1.size(2), op2.size(2)
    if batch_size_1 > batch_size_2:
        op2 = op2.repeat(1, 1, batch_size_1)[:, :, :batch_size_1]
    elif batch_size_2 > batch_size_1:
        op1 = op1.repeat(1, 1, batch_size_2)[:, :, :batch_size_2]
    kron_product = torch.einsum("ikb,jlb->ijklb", op1.contiguous(), op2.contiguous())
    return kron_product.reshape(
        op1.size(0) * op2.size(0), op1.size(1) * op2.size(1), op1.size(2)
    )


def expand_operator(
    operator: Tensor,
    qubit_support: tuple[int, ...],
    full_support: tuple[int, ...],
    diagonal: bool = False,
) -> Tensor:
    """
    Expands an operator acting on a given qubit_support to act on a larger full_support
    by explicitly filling in identity matrices on all remaining qubits.
    """
    full_support = tuple(sorted(full_support))
    qubit_support = tuple(sorted(qubit_support))
    if not set(qubit_support).issubset(set(full_support)):
        raise ValueError(
            "Expanding tensor operation requires a `full_support` argument "
            "larger than or equal to the `qubit_support`."
        )
    device, dtype = operator.device, operator.dtype
    if not diagonal:
        other = IMAT.clone().to(device=device, dtype=dtype).unsqueeze(2)
    else:
        other = IDIAG.clone().to(device=device, dtype=dtype).unsqueeze(1)

    for i in set(full_support) - set(qubit_support):
        qubit_support += (i,)
        operator = torch.kron(operator.contiguous(), other)
    operator = permute_basis(operator, qubit_support, inv=True, diagonal=diagonal)
    return operator


def promote_operator(operator: Tensor, target: int, n_qubits: int) -> Tensor:
    from pyqtorch.primitives import I

    """
    FIXME: Remove and replace usage with the `expand_operator` above.

    Promotes `operator` to the size of the circuit (number of qubits and batch).
    Targeting the first qubit implies target = 0, so target > n_qubits - 1.

    Arguments:
        operator: The operator tensor to be promoted.
        target: The index of the target qubit to which the operator is applied.
            Targeting the first qubit implies target = 0, so target > n_qubits - 1.
        n_qubits: Number of qubits in the circuit.

    Returns:
        Tensor: The promoted operator tensor.

    Raises:
        ValueError: If `target` is outside the valid range of qubits.
    """
    if target > n_qubits - 1:
        raise ValueError(
            "The target must be a valid qubit index within the circuit's range."
        )
    qubits = torch.arange(0, n_qubits)
    qubits = qubits[qubits != target]
    for qubit in qubits:
        operator = torch.where(
            target > qubit,
            operator_kron(I(target).tensor(), operator),
            operator_kron(operator, I(target).tensor()),
        )
    return operator


def permute_state(
    state: Tensor, qubit_support: tuple | list, inv: bool = False
) -> Tensor:
    """Takes a state tensor and permutes the qubit amplitudes
    according to the order of the qubit support.

    Args:
        state (Tensor): State to permute over.
        qubit_support (tuple): Qubit support.
        inv (bool): Applies the inverse permutation instead.

    Returns:
        Tensor: Permuted state.
    """
    if tuple(qubit_support) == tuple(sorted(qubit_support)):
        return state

    ordered_support = argsort(qubit_support)
    ranked_support = argsort(ordered_support)

    perm = list(ranked_support) + [len(qubit_support)]

    if inv:
        perm = np.argsort(perm).tolist()

    return state.permute(perm)


def permute_basis(
    operator: Tensor,
    qubit_support: tuple,
    inv: bool = False,
    diagonal: bool = False,
) -> Tensor:
    """Takes an operator tensor and permutes the rows and
    columns according to the order of the qubit support.

    Args:
        operator (Tensor): Operator to permute over.
        qubit_support (tuple): Qubit support.
        inv (bool): Applies the inverse permutation instead.

    Returns:
        Tensor: Permuted operator.
    """
    ordered_support = argsort(qubit_support)
    ranked_support = argsort(ordered_support)
    n_qubits = len(qubit_support)
    if all(a == b for a, b in zip(ranked_support, list(range(n_qubits)))):
        return operator
    batch_size = operator.size(-1)
    if not diagonal:
        operator = operator.view([2] * 2 * n_qubits + [batch_size])
        perm = list(
            tuple(ranked_support) + tuple(ranked_support + n_qubits) + (2 * n_qubits,)
        )
        if inv:
            perm = np.argsort(perm).tolist()
        return operator.permute(perm).reshape([2**n_qubits, 2**n_qubits, batch_size])
    else:
        operator = operator.view([2] * n_qubits + [batch_size])
        perm = list(tuple(ranked_support) + (n_qubits,))
        if inv:
            perm = np.argsort(perm).tolist()
        return operator.permute(perm).reshape([2**n_qubits, batch_size])


def random_dm_promotion(
    target: int, dm_input: DensityMatrix, n_qubits: int
) -> DensityMatrix:
    """
    Promotes the density matrix `dm_input` at the specified target qubit with random states.

    Args:
        target (int): The index of the target qubit where the promotion will be applied.
        dm_input (DensityMatrix): The input density matrix to promote.
        n_qubits (int): The total number of qubits in the quantum system.

    Returns:
        DensityMatrix: The density matrix after applying random promotion.

    Raises:
        ValueError: If `target` is not within the valid range [0, n_qubits - 1].
    """
    if target > n_qubits - 1:
        raise ValueError(
            "The target must be a valid qubit index within the circuit range."
        )
    rng = torch.Generator().manual_seed(n_qubits)
    if target == 0 or target == n_qubits - 1:
        state_random: Tensor = random_state(n_qubits - 1, generator=rng)
        dm_random: DensityMatrix = density_mat(state_random)
        dm_input = torch.where(
            torch.tensor(target) == 0,
            operator_kron(dm_input, dm_random),
            operator_kron(dm_random, dm_input),
        )
    else:
        state_random_1 = random_state(target, generator=rng)
        state_random_2 = random_state(n_qubits - (target + 1), generator=rng)
        dm_random_1, dm_random_2 = density_mat(state_random_1), density_mat(
            state_random_2
        )
        dm_input = operator_kron(dm_random_1, operator_kron(dm_input, dm_random_2))
    return dm_input


def dm_partial_trace(rho: DensityMatrix, keep_indices: list[int]) -> DensityMatrix:
    """
    Computes the partial trace of a density matrix for a system of several qubits with batch size.
    This function also permutes the qubits according to the order specified in keep_indices.

    Args:
        rho (DensityMatrix) : Density matrix of shape [2**n_qubits, 2**n_qubits, batch_size].
        keep_indices (list[int]): Index of the qubit subsystems to keep.

    Returns:
        DensityMatrix: Reduced density matrix after the partial trace,
        of shape [2**n_keep, 2**n_keep, batch_size].
    """
    n_qubits = int(log2((rho.shape[0])))
    batch_size = rho.shape[2]
    rho = rho.reshape(([2] * n_qubits * 2 + [batch_size]))

    rho_subscripts = array(ABC_ARRAY[: n_qubits * 2 + 1], copy=True)
    keep_subscripts = "".join(rho_subscripts[keep_indices]) + "".join(
        rho_subscripts[array(keep_indices) + n_qubits]
    )

    # Trace by equating indices
    trace_indices = delete(arange(n_qubits), keep_indices)
    rho_subscripts[trace_indices + n_qubits] = rho_subscripts[trace_indices]
    rho_subscripts = "".join(rho_subscripts)

    einsum_subscripts = (
        rho_subscripts + "->" + keep_subscripts + rho_subscripts[n_qubits * 2]
    )

    rho_reduced = torch.einsum(einsum_subscripts, rho)
    n_keep = len(keep_indices)
    return rho_reduced.reshape(2**n_keep, 2**n_keep, batch_size)


def generate_dm(n_qubits: int, batch_size: int) -> Tensor:
    """Generates a random density matrix using a real hermitian matrix"""
    density_list = []
    for _ in range(batch_size):
        rand_mat = torch.rand(2**n_qubits, 2**n_qubits) + 1j * torch.rand(
            2**n_qubits, 2**n_qubits
        )
        herm_mat = rand_mat + torch.transpose(rand_mat, 0, 1)
        density_mat = torch.matrix_exp(-herm_mat)
        density_mat = density_mat / torch.trace(density_mat)
        density_list.append(density_mat)

    return moveaxis(torch.stack(density_list), 0, 2)


def add_batch_dim(operator: Tensor, batch_size: int = 1) -> Tensor:
    """In case we have a sequence of batched parametric gates mixed with primitive gates,
    we adjust the batch_dim of the primitive gates to match."""
    return (
        operator.repeat(1, 1, batch_size)
        if operator.shape != (2, 2, batch_size)
        else operator
    )


def operator_to_sparse_diagonal(operator: Tensor) -> Tensor:
    """Convert operator to a sparse diagonal tensor.
    Arguments:
        operator: Operator to convert.
    Returns:
        A sparse tensor.
    """
    operator = torch.diag(operator)
    indices, values, size = (
        torch.nonzero(operator),
        operator[operator != 0],
        len(operator),
    )
    indices = torch.stack((indices.flatten(), indices.flatten()))
    return torch.sparse_coo_tensor(indices, values, (size, size))


def cache(func: Callable, *, maxsize: int = 1) -> Callable:
    """Cache a function returning a tensor by memoizing its most recent calls.

    This decorator extends `methodtools.lru_cache` to also cache a function on
    PyTorch grad mode status (enabled or disabled). This prevents cached tensors
    detached from the graph (for example computed within a `with torch.no_grad()`
    block) from being used by mistake by later code which requires tensors attached
    to the graph.

    By default, the cache size is `1`, which means that only the most recent call is
    cached. Use the `maxsize` keyword argument to change the maximum cache size.

    Warning:
        This decorator should only be used for PyTorch tensors.

    Args:
        func: Function returning a tensor, can take any number of arguments.

    Returns:
        Cached function.
    """
    if func is None:
        return partial(cache, maxsize=maxsize)

    # define a function cached on its arguments and also on PyTorch grad mode status
    @lru_cache(maxsize=maxsize)
    def grad_cached_func(*args: Any, grad_enabled: bool, **kwargs: Any) -> Tensor:
        return func(*args, **kwargs)

    # wrap `func` to call its modified cached version
    @wraps(func)
    def wrapper(*args: Any, **kwargs: Any) -> Tensor:
        return grad_cached_func(*args, grad_enabled=torch.is_grad_enabled(), **kwargs)

    return wrapper


def hairer_norm(x: Tensor) -> Tensor:
    """Rescaled Frobenius norm of a batched matrix.

    Args:
        x: Tensor of shape `(..., n, n)`.

    Returns:
        Tensor of shape `(...)` holding the norm of each matrix in the batch.
    """
    return torch.linalg.matrix_norm(x) / sqrt(x.size(-1) * x.size(-2))


@dataclass
class Result:
    states: Tensor


class SolverType(StrEnum):
    DP5_SE = "dp5_se"
    """Uses fifth-order Dormand-Prince Schrodinger equation solver"""

    DP5_ME = "dp5_me"
    """Uses fifth-order Dormand-Prince master equation solver"""

    KRYLOV_SE = "krylov_se"
    """Uses Krylov Schrodinger equation solver"""


def is_parametric(operation: pyq.Sequence) -> bool:
    """Check if operation is parametric.

    Args:
        operation (pyq.Sequence): checked operation

    Returns:
        bool: True if operation is parametric, False otherwise
    """

    from pyqtorch.primitives import Parametric

    res = False
    for m in operation.modules():
        if isinstance(m, (pyq.Scale, Parametric)):
            if isinstance(m.param_name, (str, pyq.ConcretizedCallable)):  # type: ignore[has-type]
                res = True
                break
    return res


def heaviside(x: Tensor, _: Any = None, slope: float = 1000.0) -> Tensor:
    """Torch autograd-compatible Heaviside function implementation.

    Args:
        x (Tensor): function argument
        _ (Any): unused argument left for signature compatibility reasons
        slope (float, optional): slope of Heaviside function (theoretically should be $infty$).
                                 Defaults to 1000.0.

    Returns:
        Tensor: function value
    """

    if x.ndim > 1:
        raise ValueError("Argument tensor must be 0-d or 1-d.")

    shape = (1, 2) if x.ndim == 0 else (len(x), 2)
    a = torch.zeros(shape)
    a[:, 0] = x
    return torch.clamp(
        slope * torch.max(a, dim=1)[0], torch.tensor(0.0), torch.tensor(1.0)
    )


def todense_tensor(x: Tensor) -> Tensor:
    """Convert a diagonal tensor to its dense representation.

    Args:
        x (Tensor): diagonal tensor

    Returns:
        Tensor: dense tensor
    """
    return torch.transpose(torch.diag_embed(torch.transpose(x, 0, 1)), 0, -1)



from __future__ import annotations

from functools import cached_property

from pyqtorch.utils import (
    qubit_support_as_tuple,
)


class Support:
    """
    Generic representation of the qubit support. For single qubit operations,
    a multiple index support indicates apply the operation for each index in the
    support.

    Both target and control lists must be ordered!

    Attributes:
       target = Index or indices where the operation is applied.
       control = Index or indices to which the operation is conditioned to.
    """

    def __init__(
        self,
        target: int | tuple[int, ...],
        control: int | tuple[int, ...] | None = None,
    ) -> None:
        self.target = qubit_support_as_tuple(target)
        self.control = qubit_support_as_tuple(control) if control is not None else ()

    @classmethod
    def target_all(cls) -> Support:
        return Support(target=())

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, Support):
            return NotImplemented

        return self.target == other.target and self.control == other.control

    def __len__(self):
        return len(self.qubits)

    def __add__(self, other: Support) -> Support:
        if not isinstance(other, Support):
            raise ValueError(f"Cannot add type '{type(other)}' to Support.")
        targets = tuple(set(self.target + other.target))
        controls = tuple(set(self.control + other.control))
        if len(set(targets + controls)) != (len(targets) + len(controls)):
            raise ValueError(
                "Cannot combine qubit supports with qubits in common between targets and controls."
            )
        return Support(targets, controls)

    @cached_property
    def qubits(self) -> tuple[int, ...]:
        return self.control + self.target

    @cached_property
    def sorted_qubits(self) -> tuple[int, ...]:
        return tuple(sorted(self.qubits))

    def __repr__(self) -> str:
        if not self.target:
            return f"{self.__class__.__name__}.target_all()"

        subspace = f"target: {self.target}"
        if self.control:
            subspace += f", control: {self.control}"

        return f"{self.__class__.__name__}({subspace})"



from __future__ import annotations

import pytest
import torch

from pyqtorch import RX, RY, ConcretizedCallable, Scale, Sequence, X
from pyqtorch.utils import heaviside, is_parametric


@pytest.mark.parametrize(
    "operation, result",
    [
        (RX(0, "x"), True),
        (RY(1, 0.5), False),
        (Scale(X(1), "y"), True),
        (Scale(X(1), 0.2), False),
        (Scale(X(1), ConcretizedCallable("mul", ["y", "x"])), True),
    ],
)
def test_is_parametric(operation: Sequence, result: bool) -> None:
    assert is_parametric(operation) == result


def test_heaviside() -> None:
    x = torch.linspace(-1, 1, 50)
    assert torch.allclose(heaviside(x), torch.heaviside(x, torch.tensor(0.0)))



from __future__ import annotations

from .analog import AnalogDepolarizing, AnalogNoise
from .digital_gates import (
    AmplitudeDamping,
    BitFlip,
    Depolarizing,
    GeneralizedAmplitudeDamping,
    Noise,
    PauliChannel,
    PhaseDamping,
    PhaseFlip,
)
from .digital_protocol import DigitalNoiseProtocol, DigitalNoiseType, _repr_noise
from .readout import CorrelatedReadoutNoise, ReadoutNoise, WhiteNoise



from __future__ import annotations

import random
from collections import Counter

import pytest
import torch

import pyqtorch as pyq
from pyqtorch.noise import CorrelatedReadoutNoise, ReadoutNoise
from pyqtorch.noise.readout import (
    WhiteNoise,
    bs_bitflip_corruption,
    create_confusion_matrices,
    create_noise_matrix,
    sample_to_matrix,
)
from pyqtorch.primitives import Primitive
from pyqtorch.utils import sample_multinomial
from pyqtorch.utils_distributions import js_divergence, js_divergence_counters


def test_noise_matrix():
    noise_mat = torch.tensor([[0.5679, 0.7676]])
    zero_prob_mat = torch.tensor(
        [[1.0, 0.0], [0.0, 1.0]],
        dtype=torch.float64,
    )
    assert torch.allclose(
        create_confusion_matrices(noise_mat, 0.1), torch.stack([zero_prob_mat] * 2)
    )


@pytest.mark.parametrize(
    "error_probability, counters, exp_corrupted_counters, n_qubits",
    [
        (
            1.0,
            [Counter({"00": 27, "01": 23, "10": 24, "11": 26})],
            [Counter({"11": 27, "10": 23, "01": 24, "00": 26})],
            2,
        ),
        (
            1.0,
            [Counter({"001": 27, "010": 23, "101": 24, "110": 26})],
            [Counter({"110": 27, "101": 23, "010": 24, "001": 26})],
            3,
        ),
    ],
)
def test_bitstring_corruption_all_bitflips(
    error_probability: float,
    counters: list,
    exp_corrupted_counters: list,
    n_qubits: int,
) -> None:
    n_shots = 100
    noise_matrix = create_noise_matrix(WhiteNoise.UNIFORM, n_shots, n_qubits)
    err_idx = torch.as_tensor(noise_matrix < error_probability)
    sample = sample_to_matrix(counters[0])
    corrupted_counters = [bs_bitflip_corruption(err_idx=err_idx, sample=sample)]
    assert sum(corrupted_counters[0].values()) == n_shots
    assert corrupted_counters == exp_corrupted_counters
    assert torch.allclose(
        torch.tensor(1.0 - js_divergence_counters(corrupted_counters[0], counters[0])),
        torch.ones(1),
        atol=1e-3,
    )


@pytest.mark.parametrize(
    "counters, n_qubits",
    [
        (
            [Counter({"00": 27, "01": 23, "10": 24, "11": 26})],
            2,
        ),
        (
            [Counter({"001": 27, "010": 23, "101": 24, "110": 26})],
            3,
        ),
    ],
)
def test_bitstring_corruption_mixed_bitflips(counters: list, n_qubits: int) -> None:
    n_shots = 100
    error_probability = random.random()
    noise_matrix = create_noise_matrix(WhiteNoise.UNIFORM, n_shots, n_qubits)
    err_idx = torch.as_tensor(noise_matrix < error_probability)
    sample = sample_to_matrix(counters[0])
    corrupted_counters = [bs_bitflip_corruption(err_idx=err_idx, sample=sample)]
    for noiseless, noisy in zip(counters, corrupted_counters):
        assert sum(noisy.values()) == n_shots
        assert js_divergence_counters(noiseless, noisy) >= 0.0


@pytest.mark.flaky(max_runs=5)
@pytest.mark.parametrize(
    "error_probability, n_shots, list_ops",
    [
        (0.1, 100, [pyq.X(0), pyq.X(1)]),
    ],
)
def test_readout_error_quantum_circuit(
    error_probability: float,
    n_shots: int,
    list_ops: list[Primitive],
) -> None:

    n_qubits = max([max(op.target) for op in list_ops]) + 1
    noiseless_qc = pyq.QuantumCircuit(n_qubits, list_ops)
    noiseless_samples = noiseless_qc.sample(n_shots=n_shots)

    readout = ReadoutNoise(n_qubits, error_probability=error_probability, seed=0)
    noisy_qc = pyq.QuantumCircuit(n_qubits, list_ops, readout)
    noisy_samples = noisy_qc.sample(n_shots=n_shots)

    for noiseless, noisy in zip(noiseless_samples, noisy_samples):
        assert sum(noiseless.values()) == sum(noisy.values()) == n_shots
        assert js_divergence_counters(noiseless, noisy) > 0.0
        assert torch.allclose(
            torch.tensor(1.0 - js_divergence_counters(noiseless, noisy)),
            torch.ones(1) - error_probability,
            atol=1e-1,
        )


def test_correlated_readout() -> None:
    n_shots = 1000
    confusion_matrix = torch.tensor(
        [
            [0.9, 0.05, 0.03, 0.02],
            [0.05, 0.85, 0.05, 0.05],
            [0.03, 0.05, 0.87, 0.05],
            [0.02, 0.05, 0.05, 0.88],
        ],
        dtype=torch.float64,
    )

    corr_readout = CorrelatedReadoutNoise(confusion_matrix, 0)
    probas = torch.tensor([[0.4, 0.3, 0.2, 0.1]], dtype=torch.double)
    out_probas = corr_readout.apply(probas, n_shots=1000)
    assert torch.allclose(
        out_probas,
        torch.tensor([[0.3830, 0.2900, 0.2060, 0.1210]], dtype=torch.float64),
        atol=1e-4,
    )


def test_readout_error_expectation() -> None:

    n_shots = 100
    rx = pyq.RX(0, param_name="theta")
    y = pyq.Y(0)
    cnot = pyq.CNOT(0, 1)
    ops = [rx, y, cnot]
    n_qubits = 2
    noiseless_qc = pyq.QuantumCircuit(n_qubits, ops)
    initstate = pyq.random_state(n_qubits)
    theta = torch.rand(1, requires_grad=True)
    obs = pyq.Observable(pyq.Z(0))

    readout = ReadoutNoise(n_qubits, error_probability=0.1, seed=0)
    noisy_qc = pyq.QuantumCircuit(n_qubits, ops, readout)

    confusion_matrix = torch.tensor(
        [
            [0.9, 0.05, 0.03, 0.02],
            [0.05, 0.85, 0.05, 0.05],
            [0.03, 0.05, 0.87, 0.05],
            [0.02, 0.05, 0.05, 0.88],
        ],
        dtype=torch.float64,
    )
    readoutCorrelated = CorrelatedReadoutNoise(confusion_matrix, seed=0)

    corr_noisy_qc = pyq.QuantumCircuit(n_qubits, ops, readoutCorrelated)
    assert not torch.allclose(
        pyq.expectation(
            noiseless_qc,
            initstate,
            {"theta": theta},
            observable=obs,
        ),
        pyq.expectation(
            noisy_qc, initstate, {"theta": theta}, observable=obs, n_shots=n_shots
        ),
    )
    assert not torch.allclose(
        pyq.expectation(
            noiseless_qc,
            initstate,
            {"theta": theta},
            observable=obs,
        ),
        pyq.expectation(
            corr_noisy_qc, initstate, {"theta": theta}, observable=obs, n_shots=n_shots
        ),
    )


@pytest.mark.flaky(max_runs=5)
def test_readout_apply_probas() -> None:
    n_qubits = 2
    n_shots = 1000
    probas = torch.tensor([[0.4, 0.3, 0.2, 0.1]], dtype=torch.double)
    readobj = ReadoutNoise(2, seed=0)
    out_probas = readobj.apply(probas, n_shots=1000)

    assert torch.allclose(torch.sum(out_probas), torch.ones(1, dtype=out_probas.dtype))
    assert torch.allclose(
        out_probas,
        torch.tensor([[0.3860, 0.2957, 0.2043, 0.1140]], dtype=torch.float64),
        atol=1e-4,
    )

    batch_sample_multinomial = torch.func.vmap(
        lambda p: sample_multinomial(
            p, n_qubits, n_shots, return_counter=False, minlength=probas.shape[-1]
        ),
        randomness="different",
    )

    p = batch_sample_multinomial(probas).squeeze(0) / n_shots
    q = batch_sample_multinomial(out_probas).squeeze(0) / n_shots

    jsd = js_divergence(p, q)
    assert jsd > 0.0

    assert torch.allclose(
        torch.ones(1) - jsd,
        torch.ones(1) - readobj.error_probability,
        atol=1e-1,
    )



from __future__ import annotations

from math import log2

from torch import Tensor, trace, vmap
from torch.nn import Module, ParameterDict

from pyqtorch.apply import operator_product
from pyqtorch.circuit import Sequence
from pyqtorch.composite import Add
from pyqtorch.embed import Embedding
from pyqtorch.primitives import Primitive
from pyqtorch.utils import DensityMatrix, inner_prod


class Observable(Add):
    """
    The Observable :math:`O` represents an operator from which
    we can extract expectation values from quantum states.

    Given an input state :math:`\\ket\\rangle`, the expectation value with :math:`O` is defined as
    :math:`\\langle\\bra|O\\ket\\rangle`

    Attributes:
        operations: List of operations.
        n_qubits: Number of qubits it is defined on.
    """

    def __init__(
        self,
        operations: list[Module] | Primitive | Sequence,
    ):
        super().__init__(operations if isinstance(operations, list) else [operations])

    def expectation(
        self,
        state: Tensor,
        values: dict[str, Tensor] | ParameterDict | None = None,
        embedding: Embedding | None = None,
    ) -> Tensor:
        """Calculate the inner product :math:`\\langle\\bra|O\\ket\\rangle`

        Arguments:
            state: Input state.
            values: Values of parameters.

        Returns:
            The expectation value.
        """
        values = values or dict()
        if isinstance(state, DensityMatrix):
            n_qubits = int(log2(state.size()[0]))
            obs_rho = operator_product(
                self.tensor(values=values, embedding=embedding),
                self.qubit_support,
                state,
                tuple(range(n_qubits)),
            )
            return vmap(trace)(obs_rho.permute(2, 0, 1)).real

        return inner_prod(state, self.forward(state, values, embedding)).real



from __future__ import annotations

from collections import Counter
from functools import partial
from logging import getLogger

import torch
from torch import Tensor

from pyqtorch.apply import apply_operator, apply_operator_dm
from pyqtorch.circuit import QuantumCircuit
from pyqtorch.differentiation import (
    AdjointExpectation,
    PSRExpectation,
    check_support_psr,
)
from pyqtorch.embed import Embedding
from pyqtorch.hamiltonians import Observable
from pyqtorch.utils import DensityMatrix, DiffMode, sample_multinomial

logger = getLogger(__name__)


def _values_processing(
    values: dict[str, Tensor] | dict[str, dict[str, Tensor]] | None = None
) -> tuple[dict[str, Tensor], dict[str, Tensor]]:
    """Process the parameter values."""
    values = values or dict()
    values_observables = values
    values_circuit = values
    val_keys = values.keys()
    if "circuit" in val_keys and "observables" in val_keys:
        values_observables = values["observables"]
        values_circuit = values["circuit"]
    return values_circuit, values_observables


def run(
    circuit: QuantumCircuit,
    state: Tensor = None,
    values: dict[str, Tensor] | None = None,
    embedding: Embedding | None = None,
) -> Tensor:
    """Sequentially apply each operation in `circuit` to an input state `state`
    given parameter values `values`, perform an optional `embedding` on `values`
    and return an output state.

    Arguments:
        circuit: A pyqtorch.QuantumCircuit instance.
        state: A torch.Tensor of shape [2, 2, ..., batch_size].
        values: A dictionary containing <'parameter_name': torch.Tensor> pairs denoting
                the current parameter values for each parameter in `circuit`.
        embedding: An optional instance of `Embedding`.

    Returns:
         A torch.Tensor of shape [2, 2, ..., batch_size].

    Example:

    ```python exec="on" source="material-block" html="1"
    from torch import rand
    from pyqtorch import QuantumCircuit, RY, random_state, run

    n_qubits = 2
    circ = QuantumCircuit(n_qubits, [RY(0, 'theta')])
    state = random_state(n_qubits)
    run(circ, state, {'theta': rand(1)})
    ```
    """
    values = values or dict()
    logger.debug(f"Running circuit {circuit} on state {state} and values {values}.")
    return circuit.run(state=state, values=values, embedding=embedding)


def sample(
    circuit: QuantumCircuit,
    state: Tensor = None,
    values: dict[str, Tensor] | None = None,
    n_shots: int = 1000,
    embedding: Embedding | None = None,
) -> list[Counter]:
    """Sample from `circuit` given an input state `state` given
    current parameter values `values`, perform an optional `embedding`
    on `values` and return a list Counter objects mapping from
    <bitstring: num_samples>.

    Arguments:
        circuit: A pyqtorch.QuantumCircuit instance.
        state: A torch.Tensor of shape [2, 2, ..., batch_size].
        values: A dictionary containing <'parameter_name': torch.Tensor> pairs
                denoting the current parameter values for each parameter in `circuit`.
        n_shots: A positive int denoting the number of requested samples.
        embedding: An optional instance of `Embedding`.

    Returns:
         A list of Counter objects containing <bitstring: num_samples> pairs.

    Example:

    ```python exec="on" source="material-block" html="1"
    from torch import rand
    from pyqtorch import random_state, sample, QuantumCircuit, RY

    n_qubits = 2
    circ = QuantumCircuit(n_qubits, [RY(0, 'theta')])
    state = random_state(n_qubits)
    sample(circ, state, {'theta': rand(1)}, n_shots=1000)[0]
    ```
    """
    values = values or dict()
    logger.debug(
        f"Sampling circuit {circuit} on state {state} and values {values} with n_shots {n_shots}."
    )
    return circuit.sample(
        state=state, values=values, n_shots=n_shots, embedding=embedding
    )


def analytical_expectation(
    circuit: QuantumCircuit,
    state: Tensor,
    observable: Observable,
    values: dict[str, Tensor] | None = None,
    embedding: Embedding | None = None,
) -> Tensor:
    """Compute the analytical expectation value.

    Given an initial state :math:`\\ket\\rangle`,
    a quantum circuit :math:`U(\\theta)`,
    the analytical expectation value with :math:`O` is defined as
    :math:`\\langle\\bra U_{\\dag}(\\theta) |O| U(\\theta) \\ket\\rangle`

    Args:
        circuit (QuantumCircuit): Quantum circuit :math:`U(\\theta)`.
        state (Tensor): Input state :math:`\\ket\\rangle`.
        observable (Observable): Observable O.
        values (dict[str, Tensor], optional): Parameter values for the
            circuit and the observable if any.
        embedding (Embedding | None, optional): An optional instance of `Embedding`.

    Returns:
        Tensor: Expectation value.
    """
    values_circuit, values_observables = _values_processing(values)
    state = run(circuit, state, values_circuit, embedding=embedding)
    return observable.expectation(state, values_observables, embedding=embedding)


def sampled_expectation(
    circuit: QuantumCircuit,
    state: Tensor,
    observable: Observable,
    values: dict[str, Tensor] | None = None,
    embedding: Embedding | None = None,
    n_shots: int = 1,
) -> Tensor:
    """Expectation value approximated via sampling.

    Given an input state :math:`\\ket\\rangle`,
    the expectation analytical value with :math:`O` is defined as
    :math:`\\langle\\bra|O\\ket\\rangle`

    Args:
        state (Tensor): Input state :math:`\\ket\\rangle`.
        observable (Observable): Observable O.
        values (dict[str, Tensor], optional): Parameter values for the observable if any.
        embedding (Embedding | None, optional): An optional instance of `Embedding`.
        n_shots: (int, optional): Number of samples to compute expectation on.

    Returns:
        Tensor: Expectation value.
    """
    values_circuit, values_observables = _values_processing(values)
    state = run(circuit, state, values_circuit, embedding=embedding)
    n_qubits = circuit.n_qubits

    # batchsize needs to be first dim for eigh
    eigvals, eigvecs = torch.linalg.eigh(
        observable.tensor(
            values=values_observables,
            embedding=embedding,
            full_support=tuple(range(n_qubits)),
        ).permute((2, 0, 1))
    )
    eigvals = eigvals.squeeze()
    if isinstance(state, DensityMatrix):
        eigvec_state_prod = apply_operator_dm(
            state,
            eigvecs.T.conj(),
            tuple(range(n_qubits)),
        )
        probs = torch.diagonal(eigvec_state_prod, dim1=0, dim2=1).real

    else:
        eigvec_state_prod = apply_operator(
            state,
            eigvecs.T.conj(),
            tuple(range(n_qubits)),
        )
        eigvec_state_prod = torch.flatten(
            eigvec_state_prod, start_dim=0, end_dim=-2
        ).t()
        probs = torch.pow(torch.abs(eigvec_state_prod), 2)
    if circuit.readout_noise is not None:
        batch_samples = circuit.readout_noise.apply(probs, n_shots)

    batch_sample_multinomial = torch.func.vmap(
        lambda p: sample_multinomial(
            p, n_qubits, n_shots, return_counter=False, minlength=probs.shape[-1]
        ),
        randomness="different",
    )
    batch_samples = batch_sample_multinomial(probs)

    normalized_samples = torch.div(
        batch_samples, torch.tensor(n_shots, dtype=probs.dtype)
    )
    normalized_samples.requires_grad = True
    expectations = torch.einsum(
        "i,ji ->j", eigvals.to(dtype=normalized_samples.dtype), normalized_samples
    )
    return expectations


def expectation(
    circuit: QuantumCircuit,
    state: Tensor = None,
    values: dict[str, Tensor] | dict[str, dict[str, Tensor]] | None = None,
    observable: Observable = None,  # type: ignore[assignment]
    diff_mode: DiffMode = DiffMode.AD,
    n_shots: int | None = None,
    embedding: Embedding | None = None,
) -> Tensor:
    """Compute the expectation value of `circuit` given a `state`,
    parameter values `values` and an `observable`
    and optionally compute gradients using diff_mode.

    Arguments:
        circuit: A pyqtorch.QuantumCircuit instance.
        state: A torch.Tensor of shape [2, 2, ..., batch_size].
        values: A dictionary containing <'parameter_name': torch.Tensor> pairs
                denoting the current parameter values for each parameter in `circuit`.
                Note it can include also values for the observables, but differentiation will
                not separate gradients.
                To do so, we should provide values as a dict of dict[str, Tensor]
                with two keys: `circuit` and `observables`.
        observable: A pyq.Observable instance.
        diff_mode: The differentiation mode.
        n_shots: Number of shots for estimating expectation values.
                    Only used with DiffMode.GPSR or DiffMode.AD.
        embedding: An optional instance of `Embedding`.

    Returns:
        An expectation value.

    Example:

    ```python exec="on" source="material-block" html="1"
    from torch import pi, ones_like, tensor
    from pyqtorch import random_state, RY, expectation, DiffMode, Observable, Add, Z, QuantumCircuit
    from torch.autograd import grad

    n_qubits = 2
    circ = QuantumCircuit(n_qubits, [RY(0, 'theta')])
    state = random_state(n_qubits)
    theta = tensor(pi, requires_grad=True)
    observable = Observable(Add([Z(i) for i in range(n_qubits)]))
    expval = expectation(circ, state, {'theta': theta}, observable, diff_mode = DiffMode.ADJOINT)
    dfdtheta= grad(expval, theta, ones_like(expval))[0]
    ```
    """
    values = values or dict()

    if embedding is not None and diff_mode != DiffMode.AD:
        raise NotImplementedError("Only diff_mode AD supports embedding")
    logger.debug(
        f"Computing expectation of circuit {circuit} on state {state}, values {values},\
          given observable {observable} and diff_mode {diff_mode}."
    )
    if observable is None:
        logger.error("Please provide an observable to compute expectation.")

    if state is None:
        state = circuit.init_state(batch_size=1)

    expectation_fn = analytical_expectation
    if n_shots is not None:
        if isinstance(n_shots, int) and n_shots > 0:
            expectation_fn = partial(sampled_expectation, n_shots=n_shots)
        else:
            logger.error("Please provide a 'n_shots' in options of type 'int'.")

    if diff_mode == DiffMode.AD:
        return expectation_fn(
            circuit,
            state,
            observable,
            values,
            embedding,
        )
    elif diff_mode == DiffMode.ADJOINT:
        return AdjointExpectation.apply(
            circuit,
            state,
            observable,
            embedding,
            values.keys(),
            *values.values(),
        )
    elif diff_mode == DiffMode.GPSR:
        check_support_psr(circuit)
        return PSRExpectation.apply(
            circuit,
            state,
            observable,
            embedding,
            expectation_fn,
            values.keys(),
            *values.values(),
        )
    else:
        logger.error(f"Requested diff_mode '{diff_mode}' not supported.")



from __future__ import annotations

from functools import reduce
from operator import add

import numpy as np
import pytest
import torch
import torch.autograd.gradcheck
from torch.nn import Module

import pyqtorch as pyq
from pyqtorch.embed import ConcretizedCallable, Embedding, cos, log, sin, sqrt
from pyqtorch.primitives import Primitive
from pyqtorch.utils import ATOL_embedding


@pytest.mark.parametrize(
    "fn", ["sin", "cos", "log", "tanh", "tan", "sin", "sqrt", "square"]
)
def test_univariate(fn: str) -> None:
    results = []
    x = np.random.uniform(0, 1)
    for engine_name in ["jax", "torch", "numpy"]:
        native_call = ConcretizedCallable(fn, ["x"], {}, engine_name)
        native_result = native_call(
            {"x": (torch.tensor(x) if engine_name == "torch" else x)}
        )
        results.append(native_result.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], ATOL_embedding
    )


@pytest.mark.parametrize("fn", ["mul", "add", "div", "sub"])
def test_multivariate(fn: str) -> None:
    results = []
    x = np.random.randn(1)
    y = np.random.randn(1)
    for engine_name in ["jax", "torch", "numpy"]:
        native_call = ConcretizedCallable(fn, ["x", "y"], {}, engine_name)
        native_result = native_call(
            {
                "x": torch.tensor(x) if engine_name == "torch" else x,
                "y": torch.tensor(y) if engine_name == "torch" else y,
            }
        )
        results.append(native_result.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )


def test_embedding() -> None:
    x = np.random.uniform(0, 1)
    theta = np.random.uniform(0, 1)
    results = []
    for engine_name in ["jax", "torch", "numpy"]:
        v_params = ["theta"]
        f_params = ["x"]
        leaf0, native_call0 = "%0", ConcretizedCallable(
            "mul", ["x", "theta"], {}, engine_name
        )
        embedding = Embedding(
            v_params,
            f_params,
            var_to_call={leaf0: native_call0},
            engine_name=engine_name,
        )
        inputs = {
            "x": (torch.tensor(x) if engine_name == "torch" else x),
            "theta": (torch.tensor(theta) if engine_name == "torch" else theta),
        }
        eval_0 = embedding.var_to_call["%0"](inputs)
        results.append(eval_0.item())
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )


def test_reembedding() -> None:
    x = np.random.uniform(0, 1)
    theta = np.random.uniform(0, 1)
    t = np.random.uniform(0, 1)
    t_reembed = np.random.uniform(0, 1)
    results = []
    reembedded_results = []
    for engine_name in ["jax", "torch", "numpy"]:
        v_params = ["theta"]
        f_params = ["x"]
        tparam = "t"
        leaf0, native_call0 = "%0", ConcretizedCallable(
            "mul", ["x", "theta"], {}, engine_name
        )
        leaf1, native_call1 = "%1", ConcretizedCallable(
            "mul", ["t", "%0"], {}, engine_name
        )

        leaf2, native_call2 = "%2", ConcretizedCallable("sin", ["%1"], {}, engine_name)
        embedding = Embedding(
            v_params,
            f_params,
            var_to_call={leaf0: native_call0, leaf1: native_call1, leaf2: native_call2},
            tparam_name=tparam,
            engine_name=engine_name,
        )
        inputs = {
            "x": (torch.tensor(x) if engine_name == "torch" else x),
            "theta": (torch.tensor(theta) if engine_name == "torch" else theta),
            "t": (torch.tensor(t) if engine_name == "torch" else t),
        }
        all_params = embedding.embed_all(inputs)
        new_tparam_val = (
            torch.tensor(t_reembed) if engine_name == "torch" else t_reembed
        )
        reembedded_params = embedding.reembed_tparam(all_params, new_tparam_val)
        results.append(all_params["%2"].item())
        reembedded_results.append(reembedded_params["%2"].item())
    assert all([p in ["%1", "%2"] for p in embedding.tracked_vars])
    assert "%0" not in embedding.tracked_vars
    assert np.allclose(results[0], results[1], atol=ATOL_embedding) and np.allclose(
        results[0], results[2], atol=ATOL_embedding
    )
    assert np.allclose(
        reembedded_results[0], reembedded_results[1], atol=ATOL_embedding
    ) and np.allclose(reembedded_results[0], reembedded_results[2], atol=ATOL_embedding)


def test_sample_run_expectation_grads_with_embedding() -> None:
    name0, fn0 = "fn0", ConcretizedCallable("sin", ["x"])
    name1, fn1 = "fn1", ConcretizedCallable("mul", ["fn0", "y"])
    name2, fn2 = "fn2", ConcretizedCallable("mul", ["fn1", 2.0])
    name3, fn3 = "fn3", ConcretizedCallable("log", ["fn2"])
    embedding = pyq.Embedding(
        vparam_names=["x"],
        fparam_names=["y"],
        var_to_call={name0: fn0, name1: fn1, name2: fn2, name3: fn3},
    )
    rx = pyq.RX(0, param_name=name0)
    cry = pyq.CRY(0, 1, param_name=name1)
    phase = pyq.PHASE(1, param_name=name2)
    ry = pyq.RY(1, param_name=name3)
    cnot = pyq.CNOT(1, 2)
    ops = [rx, cry, phase, ry, cnot]
    n_qubits = 3
    circ = pyq.QuantumCircuit(n_qubits, ops)
    obs = pyq.Observable(pyq.Z(0))

    state = pyq.zero_state(n_qubits)

    x = torch.rand(1, requires_grad=True)
    y = torch.rand(1, requires_grad=True)

    values_ad = {"x": x, "y": y}

    wf = pyq.run(circ, state, values_ad, embedding)
    samples = pyq.sample(circ, state, values_ad, 100, embedding)
    exp_ad = pyq.expectation(
        circ, state, values_ad, obs, pyq.DiffMode.AD, embedding=embedding
    )
    assert torch.autograd.gradcheck(
        lambda x, y: pyq.expectation(
            circ, state, {"x": x, "y": y}, obs, pyq.DiffMode.AD, embedding=embedding
        ),
        (x, y),
        atol=1e-1,  # torch.autograd.gradcheck is very susceptible to small numerical errors
    )


@pytest.mark.parametrize("engine", ["torch", "jax", "numpy"])
def test_move_embedding(engine: str) -> None:
    name0, fn0 = "fn0", ConcretizedCallable("sin", ["x"], engine_name=engine)
    name1, fn1 = "fn1", ConcretizedCallable("mul", ["fn0", "y"], engine_name=engine)
    name2, fn2 = "fn2", ConcretizedCallable("mul", ["fn1", 2.0], engine_name=engine)
    name3, fn3 = "fn3", ConcretizedCallable("log", ["fn2"], engine_name=engine)
    embedding = pyq.Embedding(
        vparam_names=["x"],
        fparam_names=["y"],
        var_to_call={name0: fn0, name1: fn1, name2: fn2, name3: fn3},
        engine_name=engine,
    )
    embedding.to(device="cpu", dtype=torch.float32)
    embedding.to(device="cpu")
    embedding.to(dtype=torch.float32)


def test_reembedding_forward() -> None:

    class CustomReembed(pyq.Add):
        def __init__(self, operations: list[Module] | Primitive | pyq.Sequence):
            super().__init__(operations=operations)

        def forward(
            self,
            state: torch.Tensor,
            values: dict[str, torch.Tensor] | None = None,
            embedding: Embedding | None = None,
        ) -> torch.Tensor:
            values = values or dict()
            return reduce(
                add,
                [
                    self.run(state, values, embedding.reembed_tparam(values, val))  # type: ignore[union-attr, arg-type]
                    for val in torch.linspace(0.0, 10, 10)
                ],
            )

        def run(
            self,
            state: torch.Tensor,
            values: dict[str, torch.Tensor],
            embedding: Embedding | None = None,
        ) -> torch.Tensor:
            for op in self.operations:
                state = op(state, values)
            return state

    leaf0, native_call0 = "%0", ConcretizedCallable("sin", ["t"], {}, "torch")
    gen = pyq.Scale(pyq.Z(0), leaf0)
    custom = CustomReembed(gen)
    embed = pyq.Embedding(
        vparam_names=[],
        fparam_names=[],
        var_to_call={leaf0: native_call0},
        tparam_name="t",
    )
    wf = custom(state=pyq.zero_state(2), values={"t": torch.rand(1)}, embedding=embed)
    assert not torch.any(torch.isnan(wf))


def test_get_independent_args() -> None:
    expr: ConcretizedCallable = sqrt(sin("x")) + cos("r") * (1.0 / log("z") * "y")
    assert set(expr.independent_args) == {"x", "y", "z", "r"}



from __future__ import annotations

import random

import pytest
import torch

import pyqtorch as pyq
from pyqtorch.primitives import ControlledPrimitive, Primitive
from pyqtorch.utils import (
    ATOL,
    random_state,
)


@pytest.mark.parametrize(
    "op",
    [
        pyq.X,
        pyq.Z,
        pyq.T,
        pyq.S,
        pyq.SDagger,
        pyq.Y,
        pyq.N,
    ],
)
def test_mutation_single(op: Primitive) -> None:
    # checking mutation is equivalent to the original forward method
    n_qubits = random.randint(1, 5)
    target = random.randint(0, n_qubits - 1)
    state = random_state(n_qubits)
    gate = op(target)

    primitive_op = Primitive(gate.operation, qubit_support=gate.target)
    assert torch.allclose(gate(state), primitive_op(state), atol=ATOL)


@pytest.mark.parametrize(
    "op, op_str",
    [
        (pyq.CNOT, "X"),
        (pyq.CZ, "Z"),
        (pyq.Toffoli, "X"),
    ],
)
def test_mutation_controlled(op: Primitive, op_str: str) -> None:
    # checking mutation is equivalent to the original forward method
    n_qubits = random.randint(3, 6)
    target = random.randint(0, n_qubits - 1)
    control = random.choice([i for i in range(n_qubits) if i != target])
    if op == pyq.Toffoli:
        control = (  # type: ignore[assignment]
            control,
            random.choice([i for i in range(n_qubits) if i not in (target, control)]),
        )
    state = random_state(n_qubits)
    gate = op(control, target)

    primitive_op = ControlledPrimitive(op_str, control=gate.control, target=gate.target)
    assert torch.allclose(gate(state), primitive_op(state), atol=ATOL)


def test_mutation_swap() -> None:
    # checking mutation is equivalent to the original forward method
    n_qubits = random.randint(2, 5)
    target = random.randint(0, n_qubits - 1)
    target2 = random.choice([i for i in range(n_qubits) if i != target])
    gate = pyq.SWAP(target, target2)
    state = random_state(n_qubits)
    primitive_op = Primitive(gate.operation, qubit_support=gate.target)
    assert torch.allclose(gate(state), primitive_op(state), atol=ATOL)



from __future__ import annotations

import math
from collections import Counter

import torch
from torch import Tensor


def shannon_entropy(counter: Counter) -> float:
    return float(-sum([count * math.log(count) for count in counter.values()]))


def js_divergence_counters(counter_p: Counter, counter_q: Counter) -> float:
    """
    Compute the Jensen-Shannon divergence between two probability distributions.

    represented as Counter objects.
    The JSD is calculated using only the shared keys between the two input Counter objects.

    Args:
        counter_p (Counter): Counter of bitstring counts for probability mass function P.
        counter_q (Counter): Counter of bitstring counts for probability mass function Q.

    Returns:
        float: The Jensen-Shannon divergence between counter_p and counter_q.
    """
    # Normalise counters
    normalisation_p = sum([count for count in counter_p.values()])
    normalisation_q = sum([count for count in counter_q.values()])
    counter_p = Counter({k: v / normalisation_p for k, v in counter_p.items()})
    counter_q = Counter({k: v / normalisation_q for k, v in counter_q.items()})

    average_proba_counter = counter_p + counter_q
    average_proba_counter = Counter(
        {k: v / 2.0 for k, v in average_proba_counter.items()}
    )
    average_entropy = shannon_entropy(average_proba_counter)

    entropy_p = shannon_entropy(counter_p)
    entropy_q = shannon_entropy(counter_q)
    return float(average_entropy - (entropy_p + entropy_q) / 2.0)


def js_divergence(
    proba_mass_P: Tensor, proba_mass_Q: Tensor, epsilon: float = 1e-6
) -> Tensor:
    """Compute the Jensen-Shannon divergence between two probability distributions.

    Args:
        proba_mass_P (Tensor): Probability mass function P
        proba_mass_Q (Tensor): Probability mass function Q
        epsilon (float, optional): Small number to avoid 0 division. Defaults to 1e-6.

    Returns:
        Tensor: The Jensen-Shannon divergence between P and Q.
    """

    # Clamp values to avoid log(0)
    proba_mass_P = torch.clamp(proba_mass_P, min=epsilon)
    proba_mass_Q = torch.clamp(proba_mass_Q, min=epsilon)

    # Calculate the middle point distribution
    m = 0.5 * (proba_mass_P + proba_mass_Q)

    # Calculate KL divergence for both distributions with respect to m
    kl_p_m = torch.sum(proba_mass_P * torch.log2(proba_mass_P / m), dim=-1)
    kl_q_m = torch.sum(proba_mass_Q * torch.log2(proba_mass_Q / m), dim=-1)

    # Jensen-Shannon divergence is the average of the two KL divergences
    return 0.5 * (kl_p_m + kl_q_m)



from __future__ import annotations

from math import log2
from string import ascii_letters as ABC

from numpy import array
from numpy.typing import NDArray
from torch import Tensor, einsum

from pyqtorch.matrices import _dagger
from pyqtorch.utils import DensityMatrix, permute_basis, permute_state

ABC_ARRAY: NDArray = array(list(ABC))


def apply_operator(
    state: Tensor,
    operator: Tensor,
    qubit_support: tuple[int, ...] | list[int],
    diagonal: bool = False,
) -> Tensor:
    """Applies an operator, i.e. a single tensor of shape [2, 2, ...], on a given state
       of shape [2 for _ in range(n_qubits)] for a given set of (target and control) qubits.

       Since dimension 'i' in 'state' corresponds to all amplitudes where qubit 'i' is 1,
       target and control qubits represent the dimensions over which to contract the 'operator'.
       Contraction means applying the 'dot' operation between the operator array and dimension 'i'
       of 'state, resulting in a new state where the result of the 'dot' operation has been moved to
       dimension 'i' of 'state'. To restore the former order of dimensions, the affected dimensions
       are moved to their original positions and the state is returned.

    Arguments:
        state: State to operate on.
        operator: Tensor to contract over 'state'.
        qubit_support: Tuple of qubits on which to apply the 'operator' to.
        diagonal: Whether operator is diagonal or not.

    Returns:
        State after applying 'operator'.
    """
    qubit_support = list(qubit_support)
    n_qubits = len(state.size()) - 1
    n_support = len(qubit_support)
    n_state_dims = n_qubits + 1

    in_state_dims = ABC_ARRAY[0:n_state_dims].copy()
    if not diagonal:
        operator = operator.view([2] * n_support * 2 + [operator.size(-1)])
        operator_dims = ABC_ARRAY[
            n_state_dims : n_state_dims + 2 * n_support + 1
        ].copy()
        operator_dims[n_support : 2 * n_support] = in_state_dims[qubit_support]
    else:
        operator = operator.view([2] * n_support + [operator.size(-1)])
        operator_dims = ABC_ARRAY[n_state_dims : n_state_dims + n_support + 1].copy()
        operator_dims[:n_support] = in_state_dims[qubit_support]

    operator_dims[-1] = in_state_dims[-1]
    out_state_dims = in_state_dims.copy()
    out_state_dims[qubit_support] = operator_dims[0:n_support]
    operator_dims, in_state_dims, out_state_dims = list(
        map(lambda e: "".join(list(e)), [operator_dims, in_state_dims, out_state_dims])
    )
    return einsum(f"{operator_dims},{in_state_dims}->{out_state_dims}", operator, state)


def apply_operator_permute(
    state: Tensor,
    operator: Tensor,
    qubit_support: tuple[int, ...] | list[int],
    diagonal: bool = False,
) -> Tensor:
    """NOTE: Currently not being used.

       Alternative apply operator function with a logic based on state permutations.
       Seems to be as fast as the current `apply_operator`. To be saved for now, we
       may want to switch to this one in the future if we wish to remove the state
       [2] * n_qubits shape and make the batch dimension the first one as the typical
       torch convention.

    Arguments:
        state: State to operate on.
        operator: Tensor to contract over 'state'.
        qubit_support: Tuple of qubits on which to apply the 'operator' to.
        diagonal: Whether operator is diagonal or not.

    Returns:
        State after applying 'operator'.
    """
    n_qubits = len(state.size()) - 1
    n_support = len(qubit_support)
    batch_size = max(state.size(-1), operator.size(-1))
    full_support = tuple(range(n_qubits))
    support_perm = list(sorted(qubit_support)) + list(
        set(full_support) - set(qubit_support)
    )
    state = permute_state(state, support_perm)
    state = state.reshape([2**n_support, 2 ** (n_qubits - n_support), state.size(-1)])
    if not diagonal:
        einsum_expr = "ijb,jkb->ikb"
    else:
        einsum_expr = "jb,jkb->jkb"
    result = einsum(einsum_expr, operator, state).reshape([2] * n_qubits + [batch_size])
    return permute_state(result, support_perm, inv=True)


def apply_operator_dm(
    state: DensityMatrix,
    operator: Tensor,
    qubit_support: tuple[int, ...] | list[int],
    diagonal: bool = False,
) -> Tensor:
    """
    Apply an operator to a density matrix on a given qubit suport, i.e., compute:

    OP.DM.OP.dagger()

    Args:
        state: State to operate on.
        operator: Tensor to contract over 'state'.
        qubit_support: Tuple of qubits on which to apply the 'operator' to.
        diagonal: Whether operator is diagonal or not.

    Returns:
        DensityMatrix: The resulting density matrix after applying the operator.
    """

    if not isinstance(state, DensityMatrix):
        raise TypeError("Function apply_operator_dm requires a density matrix state.")

    n_qubits = int(log2(state.size()[0]))
    n_support = len(qubit_support)
    batch_size = max(state.size(-1), operator.size(-1))
    full_support = tuple(range(n_qubits))
    support_perm = tuple(sorted(qubit_support)) + tuple(
        set(full_support) - set(qubit_support)
    )
    state = permute_basis(state, support_perm)

    # There is probably a smart way to represent the lines below in a single einsum...
    if not diagonal:
        einsum_expr = "ijb,jkb->ikb"
    else:
        einsum_expr = "jb,jkb->jkb"
    state = state.reshape(
        [2**n_support, (2 ** (2 * n_qubits - n_support)), state.size(-1)]
    )
    state = einsum(einsum_expr, operator, state).reshape(
        [2**n_qubits, 2**n_qubits, batch_size]
    )
    state = _dagger(state).reshape(
        [2**n_support, (2 ** (2 * n_qubits - n_support)), state.size(-1)]
    )
    state = _dagger(
        einsum(einsum_expr, operator, state).reshape(
            [2**n_qubits, 2**n_qubits, batch_size]
        )
    )
    return permute_basis(state, support_perm, inv=True)


def operator_product(
    op1: Tensor,
    supp1: tuple[int, ...],
    op2: Tensor,
    supp2: tuple[int, ...],
) -> Tensor:
    """
    Operator product based on block matrix multiplication.

    E.g., for some small operator S and a big operator with 4 partitions A, B, C, D:

    |S 0|.|A B| = |S.A S.B|
    |0 S| |C D|   |S.C S.D|

    or

    |A B|.|S 0| = |A.S B.S|
    |C D|.|0 S|   |C.S D.S|

    The same generalizes for different sizes of the big operator. Note that the block
    diagonal matrix is never computed. Instead, the big operator is permuted and
    reshaped into a wide matrix:

    W = [A B C D]

    And then the result is computed as S.W, reshaped back into a square matrix, and
    permuted back into the original ordering.
    """

    if supp1 == supp2:
        return einsum("ijb,jkb->ikb", op1, op2)

    if len(supp1) < len(supp2):
        small_op, small_supp = op1, supp1
        big_op, big_supp = op2, supp2
        transpose = False
    else:
        small_op, small_supp = _dagger(op2), supp2
        big_op, big_supp = _dagger(op1), supp1
        transpose = True

    if not set(small_supp).issubset(set(big_supp)):
        raise ValueError("Operator product requires overlapping qubit supports.")

    n_big, n_small = len(big_supp), len(small_supp)
    batch_big, batch_small = big_op.size(-1), small_op.size(-1)
    batch_size = max(batch_big, batch_small)

    # Permute the large operator and reshape into a wide matrix
    support_perm = tuple(sorted(small_supp)) + tuple(set(big_supp) - set(small_supp))
    big_op = permute_basis(big_op, support_perm)
    big_op = big_op.reshape([2**n_small, (2 ** (2 * n_big - n_small)), batch_big])

    # Compute S.W and reshape back to square
    result = einsum("ijb,jkb->ikb", small_op, big_op).reshape(
        [2**n_big, 2**n_big, batch_size]
    )

    # Apply the inverse qubit permutation
    if transpose:
        return _dagger(permute_basis(result, support_perm, inv=True))
    else:
        return permute_basis(result, support_perm, inv=True)



from __future__ import annotations

from collections import Counter
from functools import reduce
from logging import getLogger
from operator import add

import torch
from torch import Tensor, bernoulli, tensor
from torch.nn import Module, ParameterDict

from pyqtorch.composite import Sequence
from pyqtorch.embed import Embedding
from pyqtorch.noise.readout import ReadoutInterface as Readout
from pyqtorch.utils import (
    DensityMatrix,
    DropoutMode,
    State,
    product_state,
    sample_multinomial,
    zero_state,
)

logger = getLogger(__name__)


class QuantumCircuit(Sequence):
    """A QuantumCircuit defining a register / number of qubits of the full system.

    Attributes:
        n_qubits (int): Number of qubits.
        operations (list[Module]): List of operations.
        readout_noise (Readout | None, optional): Readout noise
            applied to samples. Defaults to None.
    """

    def __init__(
        self,
        n_qubits: int,
        operations: list[Module],
        readout_noise: Readout | None = None,
    ):
        """Initializes QuantumCircuit.

        Args:
            n_qubits (int): Number of qubits.
            operations (list[Module]): List of operations.
            readout_noise (ReadoutNoise | None, optional): Readout noise
                applied to samples. Defaults to None.
        """
        super().__init__(operations)
        self.n_qubits = n_qubits
        self.readout_noise = readout_noise

    def run(
        self,
        state: Tensor = None,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        if state is None:
            state = self.init_state()
        elif isinstance(state, str):
            state = self.state_from_bitstring(state)
        return self.forward(state, values, embedding)

    def __hash__(self) -> int:
        return hash(reduce(add, (hash(op) for op in self.operations))) + hash(
            self.n_qubits
        )

    def init_state(self, batch_size: int = 1) -> Tensor:
        return zero_state(
            self.n_qubits, batch_size, device=self.device, dtype=self.dtype
        )

    def state_from_bitstring(self, bitstring: str, batch_size: int = 1) -> Tensor:
        return product_state(bitstring, batch_size, self.device, self.dtype)

    def sample(
        self,
        state: Tensor = None,
        values: dict[str, Tensor] | None = None,
        n_shots: int = 1000,
        embedding: Embedding | None = None,
    ) -> list[Counter]:
        values = values or dict()
        if n_shots < 1:
            raise ValueError(
                f"You can only call sample with a non-negative value for `n_shots`. Got {n_shots}."
            )

        with torch.no_grad():
            state = self.run(state=state, values=values, embedding=embedding)
            if isinstance(state, DensityMatrix):
                probs = torch.diagonal(state, dim1=0, dim2=1).real
            else:
                state = torch.flatten(
                    state,
                    start_dim=0,
                    end_dim=-2,
                ).t()
                probs = torch.pow(torch.abs(state), 2)

            if self.readout_noise is not None:
                probs = self.readout_noise.apply(probs, n_shots)
            counters = list(
                map(lambda p: sample_multinomial(p, self.n_qubits, n_shots), probs)
            )
            return counters


class DropoutQuantumCircuit(QuantumCircuit):
    """Creates a quantum circuit able to perform quantum dropout, based on the work of https://arxiv.org/abs/2310.04120.
    Args:
        dropout_mode (DropoutMode): type of dropout to perform. Defaults to DropoutMode.ROTATIONAL
        dropout_prob (float): dropout probability. Defaults to 0.06.
    """

    def __init__(
        self,
        n_qubits: int,
        operations: list[Module],
        readout_noise: Readout | None = None,
        dropout_mode: DropoutMode = DropoutMode.ROTATIONAL,
        dropout_prob: float = 0.06,
    ):
        super().__init__(n_qubits, operations, readout_noise)
        self.dropout_mode = dropout_mode
        self.dropout_prob = dropout_prob

        self.dropout_fn = getattr(self, dropout_mode)

    def forward(
        self,
        state: State,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        if self.training:
            state = self.dropout_fn(state, values)
        else:
            for op in self.operations:
                state = op(state, values, embedding)
        return state

    def rotational_dropout(
        self,
        state: State = None,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        """Randomly drops entangling rotational gates.

        Args:
            state (State, optional): pure state vector . Defaults to None.
            values (dict[str, Tensor] | ParameterDict, optional): gate parameters. Defaults to {}.

        Returns:
            State: pure state vector
        """
        for op in self.operations:
            if not (
                (hasattr(op, "param_name"))
                and (values[op.param_name].requires_grad)
                and not (int(1 - bernoulli(tensor(self.dropout_prob))))
            ):
                state = op(state, values, embedding)

        return state

    def entangling_dropout(
        self,
        state: State = None,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        """Randomly drops entangling gates.

        Args:
            state (State, optional): pure state vector. Defaults to None.
            values (dict[str, Tensor] | ParameterDict, optional): gate parameters. Defaults to {}.

        Returns:
            State: pure state vector
        """
        for op in self.operations:
            has_param = hasattr(op, "param_name")
            keep = int(1 - bernoulli(tensor(self.dropout_prob)))

            if has_param or keep:
                state = op(state, values, embedding)

        return state

    def canonical_fwd_dropout(
        self,
        state: State = None,
        values: dict[str, Tensor] | ParameterDict = dict(),
        embedding: Embedding | None = None,
    ) -> State:
        """Randomly drops rotational gates and next immediate entangling
        gates whose target bit is located on dropped rotational gates.

        Args:
            state (State, optional): pure state vector. Defaults to None.
            values (dict[str, Tensor] | ParameterDict, optional): gate parameters. Defaults to {}.

        Returns:
            State: pure state vector
        """
        entanglers_to_drop = dict.fromkeys(range(state.ndim - 1), 0)  # type: ignore
        for op in self.operations:
            if (
                hasattr(op, "param_name")
                and (values[op.param_name].requires_grad)
                and not (int(1 - bernoulli(tensor(self.dropout_prob))))
            ):
                entanglers_to_drop[op.target] = 1
            else:
                if not hasattr(op, "param_name") and (
                    entanglers_to_drop[op.control[0]] == 1
                ):
                    entanglers_to_drop[op.control[0]] = 0
                else:
                    state = op(state, values, embedding)

        return state



# # Licensed under the Apache License, Version 2.0 (the "License");
# # you may not use this file except in compliance with the License.
# # You may obtain a copy of the License at
# #     http://www.apache.org/licenses/LICENSE-2.0
# # Unless required by applicable law or agreed to in writing, software
# # distributed under the License is distributed on an "AS IS" BASIS,
# # WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# # See the License for the specific language governing permissions and
# # limitations under the License.
from __future__ import annotations

import logging
import os
import sys

logging_levels = {
    "DEBUG": logging.DEBUG,
    "INFO": logging.INFO,
    "WARNING": logging.WARNING,
    "ERROR": logging.ERROR,
    "CRITICAL": logging.CRITICAL,
}

LOG_LEVEL: str = os.environ.get("PYQ_LOG_LEVEL", "").upper()


# if LOG_LEVEL:
LOG_LEVEL: int = logging_levels.get(LOG_LEVEL, logging.INFO)  # type: ignore[arg-type, no-redef]
# If logger not setup, add handler to stderr
# else use setup presumably from Qadence
handle = None
if __name__ not in logging.Logger.manager.loggerDict.keys():
    handle = logging.StreamHandler(sys.stderr)
    handle.set_name("console")

logger = logging.getLogger(__name__)
if handle:
    logger.addHandler(handle)
[
    h.setLevel(LOG_LEVEL)  # type: ignore[func-returns-value]
    for h in logger.handlers
    if h.get_name() == "console"
]
logger.setLevel(LOG_LEVEL)

from .api import expectation, run, sample
from .apply import apply_operator
from .circuit import DropoutQuantumCircuit, QuantumCircuit
from .composite import (
    Add,
    Merge,
    Scale,
    Sequence,
)
from .embed import (
    ConcretizedCallable,
    Embedding,
    cos,
    log,
    sin,
    sqrt,
    tan,
    tanh,
)
from .hamiltonians import HamiltonianEvolution, Observable
from .noise import (
    AmplitudeDamping,
    BitFlip,
    Depolarizing,
    GeneralizedAmplitudeDamping,
    PauliChannel,
    PhaseDamping,
    PhaseFlip,
)
from .primitives import (
    CNOT,
    CPHASE,
    CRX,
    CRY,
    CRZ,
    CSWAP,
    CY,
    CZ,
    OPS_1Q,
    OPS_2Q,
    OPS_3Q,
    OPS_DIGITAL,
    OPS_PARAM,
    OPS_PARAM_1Q,
    OPS_PARAM_2Q,
    OPS_PAULI,
    PHASE,
    RX,
    RY,
    RZ,
    SWAP,
    H,
    I,
    N,
    Projector,
    S,
    SDagger,
    T,
    Toffoli,
    U,
    X,
    Y,
    Z,
)
from .time_dependent.mesolve import mesolve
from .time_dependent.sesolve import sesolve
from .utils import (
    DEFAULT_MATRIX_DTYPE,
    DEFAULT_REAL_DTYPE,
    DiffMode,
    density_mat,
    inner_prod,
    is_normalized,
    overlap,
    product_state,
    random_state,
    uniform_state,
    zero_state,
)

__all__ = [
    "ConcretizedCallable",
    "Embedding",
    "run",
    "sample",
    "expectation",
    "Add",
    "HamiltonianEvolution",
    "Scale",
    "Merge",
    "QuantumCircuit",
    "DropoutQuantumCircuit",
    "Sequence",
    "CPHASE",
    "CRX",
    "CRY",
    "CRZ",
    "PHASE",
    "RX",
    "RY",
    "RZ",
    "U",
    "DEFAULT_MATRIX_DTYPE",
    "DEFAULT_REAL_DTYPE",
    "DiffMode",
    "inner_prod",
    "is_normalized",
    "overlap",
    "random_state",
    "uniform_state",
    "zero_state",
    "CNOT",
    "CSWAP",
    "CY",
    "CZ",
    "SWAP",
    "H",
    "I",
    "N",
    "Projector",
    "S",
    "SDagger",
    "T",
    "Toffoli",
    "X",
    "Y",
    "Z",
    "apply_operator",
    "Observable",
    "sesolve",
    "mesolve",
]



from __future__ import annotations

import random

import pytest
import torch

import pyqtorch as pyq
from pyqtorch import run, sample
from pyqtorch.noise import DigitalNoiseProtocol, DigitalNoiseType
from pyqtorch.utils import (
    DensityMatrix,
    density_mat,
    product_state,
    todense_tensor,
)


def test_device_inference() -> None:
    ops = [pyq.RX(0), pyq.RX(0)]
    circ = pyq.QuantumCircuit(2, ops)
    nested_circ = pyq.QuantumCircuit(2, [circ, circ])
    assert nested_circ._device is not None


@pytest.mark.parametrize("fn", [pyq.X, pyq.Z, pyq.Y])
def test_scale(fn: pyq.primitives.Primitive) -> None:
    n_qubits = torch.randint(low=1, high=4, size=(1,)).item()
    target = random.choice([i for i in range(n_qubits)])
    state = pyq.random_state(n_qubits)
    gate = fn(target)
    values = {"scale": torch.rand(1)}
    wf = values["scale"] * pyq.QuantumCircuit(2, [gate])(state, {})

    scale_primitive = pyq.Scale(gate, "scale")
    scale_composite = pyq.Scale(pyq.Sequence([gate]), "scale")

    if gate.is_diagonal:
        assert scale_composite.is_diagonal
        assert scale_primitive.is_diagonal

    scaledwf_primitive = scale_primitive(state, values)
    scaledwf_composite = scale_composite(state, values)
    assert torch.allclose(wf, scaledwf_primitive)
    assert torch.allclose(wf, scaledwf_composite)


def test_add() -> None:
    num_gates = 2
    fns = [pyq.X, pyq.Y, pyq.Z, pyq.S, pyq.H, pyq.T]
    ops = []
    n_qubits = torch.randint(low=1, high=4, size=(1,)).item()
    state = pyq.random_state(n_qubits)
    chosen_gate_ids = torch.randint(0, len(fns) - 1, size=(num_gates,))
    for id in chosen_gate_ids:
        target = random.choice([i for i in range(n_qubits)])
        ops.append(fns[id](target))

    assert torch.allclose(pyq.Add(ops)(state), ops[0](state) + ops[1](state))

    # test diagonal ops
    diagonal_add = pyq.Add(
        [pyq.Z(random.choice([i for i in range(n_qubits)])) for _ in range(num_gates)]
    )
    assert diagonal_add.is_diagonal
    tensor_add = diagonal_add.tensor()
    tensor_add_diagonal = diagonal_add.tensor(diagonal=True)
    dense_diagonal = todense_tensor(tensor_add_diagonal)
    assert torch.allclose(tensor_add, dense_diagonal)


def test_add_commute() -> None:
    op = pyq.Add([pyq.Z(0)])
    assert not op.commuting_terms()

    op = pyq.Add([pyq.Z(q) for q in range(5)])
    assert op.commuting_terms()

    op = pyq.Add([pyq.Merge([pyq.Z(0), pyq.X(0)]), pyq.Z(1)])
    assert not op._symplectic_commute()
    assert op.commuting_terms()

    op = pyq.Add([pyq.Sequence([pyq.Z(0), pyq.X(0)]), pyq.Z(1)])
    assert not op._symplectic_commute()
    assert op.commuting_terms()


def test_merge() -> None:
    ops = [pyq.RX(0, "theta_0"), pyq.RY(0, "theta_1"), pyq.RX(0, "theta_2")]
    circ = pyq.QuantumCircuit(2, ops)
    mergecirc = pyq.Merge(ops)
    state = pyq.random_state(2)
    values = {f"theta_{i}": torch.rand(1) for i in range(3)}
    assert torch.allclose(circ(state, values), mergecirc(state, values))

    # test with density matrices
    state = density_mat(state)
    circ_out = circ(state, values)
    mergecirc_out = mergecirc(state, values)
    assert isinstance(circ_out, DensityMatrix)
    assert isinstance(mergecirc_out, DensityMatrix)
    assert torch.allclose(circ_out, mergecirc_out)

    # test diagonal merge
    ops = [pyq.RZ(0, "theta_0"), pyq.RZ(0, "theta_1")]
    circ = pyq.QuantumCircuit(2, ops)
    mergecirc = pyq.Merge(ops)
    assert mergecirc.is_diagonal
    tensor_merge = mergecirc.tensor(values)
    tensor_merge_diagonal = mergecirc.tensor(values, diagonal=True)
    dense_diagonal = todense_tensor(tensor_merge_diagonal)
    assert torch.allclose(tensor_merge, dense_diagonal)


def test_merge_noisy_op() -> None:
    ops = [
        pyq.RX(0, "theta_0"),
        pyq.RY(
            0,
            "theta_1",
            noise=DigitalNoiseProtocol(DigitalNoiseType.DEPOLARIZING, 0.1, 0),
        ),
        pyq.RX(0, "theta_2"),
    ]
    circ = pyq.QuantumCircuit(2, ops)
    mergecirc = pyq.Merge(ops)
    state = pyq.random_state(2)
    values = {f"theta_{i}": torch.rand(1) for i in range(3)}
    circ_out = circ(state, values)
    mergecirc_out = mergecirc(state, values)
    assert isinstance(circ_out, DensityMatrix)
    assert isinstance(mergecirc_out, DensityMatrix)
    assert torch.allclose(circ_out, mergecirc_out)


@pytest.mark.xfail(
    reason="Can only merge single qubit gates acting on the same qubit support."
)
def test_merge_different_sup_expect_fail() -> None:
    ops = [pyq.RX(0, "theta_0"), pyq.RY(1, "theta_1")]
    pyq.Merge(ops)


@pytest.mark.xfail(
    reason="Can only merge single qubit gates acting on the same qubit support."
)
def test_merge_multiqubit_expect_fail() -> None:
    ops = [pyq.CNOT(0, 1), pyq.RY(1, "theta_1")]
    pyq.Merge(ops)


@pytest.mark.parametrize("batch_size", [1, 2, 3])
def test_merge_different_batchsize(batch_size: int) -> None:
    ops = [pyq.X(0), pyq.RX(0, "theta_0")]
    mergecirc = pyq.Merge(ops)
    for bs in [1, batch_size]:
        mergecirc(pyq.random_state(2, bs), {"theta_0": torch.rand(batch_size)})
        mergecirc(pyq.random_state(2, batch_size), {"theta_0": torch.rand(bs)})


def test_merge_nested_dict() -> None:
    ops = [pyq.X(0), pyq.RX(0, "theta_0")]
    mergecirc = pyq.Merge(ops)
    vals = {
        "theta_0": torch.rand(1),
        "theta_2": torch.rand(2),
        "theta_3": torch.rand(2),
    }
    vals["nested"] = vals
    mergecirc(pyq.random_state(2), vals)


def test_sample_run() -> None:
    ops = [pyq.X(0), pyq.X(1)]
    circ = pyq.QuantumCircuit(4, ops)
    wf = run(circ)
    samples = sample(circ)
    assert torch.allclose(wf, product_state("1100"))
    assert torch.allclose(pyq.QuantumCircuit(4, [pyq.I(0)]).run("1100"), wf)
    assert "1100" in samples[0]



from __future__ import annotations

from logging import getLogger
from typing import Any, Tuple

from torch import Tensor, no_grad
from torch.autograd import Function

from pyqtorch.apply import apply_operator
from pyqtorch.circuit import QuantumCircuit
from pyqtorch.composite import Scale
from pyqtorch.embed import Embedding
from pyqtorch.hamiltonians import HamiltonianEvolution, Observable
from pyqtorch.primitives import Parametric, Primitive
from pyqtorch.utils import inner_prod, param_dict

logger = getLogger(__name__)


class AdjointExpectation(Function):
    """
    The adjoint differentiation method (https://arxiv.org/pdf/2009.02823.pdf) implemented as a
    custom torch.autograd.Function. It is able to perform a backward pass in O(P) time
    and maintaining atmost 3 states where P is the number of parameters in a variational circuit.

    Arguments:
        circuit: A QuantumCircuit instance
        observable: A hamiltonian.
        state: A state in the form of [2 * n_qubits + [batch_size]]
        param_names: A list of parameter names.
        *param_values: A unpacked tensor of values for each parameter.

    The forward method expects each of the above arguments, computes the expectation value
    and stores all of the above arguments in the 'ctx' (context) object along with the
    the state after applying 'circuit', 'out_state', and the 'projected_state', i.e. after applying
    'observable' to 'out_state'.

    In the 'backward', the circuit is traversed in reverse order,
    starting from the last gate to the first gate, 'undoing' each gate using the dagger operation on
    both 'out_state' and 'projected_state'.

    In case of a parametric gate (with a parameter which requires_grad):
    (1) We compute the jacobian of the operator and apply it to 'out_state'
    (2) Compute the inner product with 'projected_state' which yields the gradient.
    """

    @staticmethod
    @no_grad()
    def forward(
        ctx: Any,
        circuit: QuantumCircuit,
        state: Tensor,
        observable: Observable,
        embedding: Embedding | None,
        param_names: list[str],
        *param_values: Tensor,
    ) -> Tensor:
        ctx.circuit = circuit
        ctx.observable = observable
        ctx.embedding = embedding
        ctx.param_names = param_names
        values = param_dict(param_names, param_values)
        if embedding is not None:
            msg = "AdjointExpectation does not support Embedding."
            logger.error(msg)
            raise NotImplementedError(msg)
        ctx.out_state = circuit.run(state, values, embedding)
        ctx.projected_state = observable(ctx.out_state, values)
        ctx.save_for_backward(*param_values)
        return inner_prod(ctx.out_state, ctx.projected_state).real

    @staticmethod
    @no_grad()
    def backward(ctx: Any, grad_out: Tensor) -> Tuple[None, ...]:
        param_values = ctx.saved_tensors
        values = param_dict(ctx.param_names, param_values)
        grads_dict = {k: None for k in values.keys()}
        for op in ctx.circuit.flatten()[::-1]:
            if isinstance(op, (Primitive, Parametric, HamiltonianEvolution)):
                ctx.out_state = apply_operator(
                    ctx.out_state, op.dagger(values, ctx.embedding), op.qubit_support
                )
                if (
                    isinstance(op, (Parametric, HamiltonianEvolution))
                    and op.is_parametric
                ):
                    param_name: str = op.param_name  # type: ignore[assignment]
                    if values[param_name].requires_grad:
                        mu = apply_operator(
                            ctx.out_state,
                            op.jacobian(values, ctx.embedding),
                            op.qubit_support,
                        )
                        grad = grad_out * 2 * inner_prod(ctx.projected_state, mu).real
                    if grads_dict[param_name] is not None:
                        grads_dict[param_name] += grad
                    else:
                        grads_dict[param_name] = grad

                ctx.projected_state = apply_operator(
                    ctx.projected_state,
                    op.dagger(values, ctx.embedding),
                    op.qubit_support,
                )

            elif isinstance(op, Scale):
                # TODO: Properly fix or not support Scale in adjoint at all
                logger.error(
                    f"AdjointExpectation does not support operation: {type(op)}."
                )
                if not len(op.operations) == 1 and isinstance(
                    op.operations[0], Primitive
                ):
                    logger.error(
                        "Adjoint can only be used on Scale with Primitive blocks."
                    )
                ctx.out_state = apply_operator(
                    ctx.out_state, op.dagger(values, ctx.embedding), op.qubit_support
                )
                scaled_pyq_op = op.operations[0]
                if (
                    isinstance(scaled_pyq_op, Parametric)
                    and isinstance(scaled_pyq_op.param_name, str)
                    and values[scaled_pyq_op.param_name].requires_grad
                ):
                    mu = apply_operator(
                        ctx.out_state,
                        scaled_pyq_op.jacobian(values, ctx.embedding),
                        scaled_pyq_op.qubit_support,
                    )
                    grads_dict[scaled_pyq_op.param_name] = (
                        grad_out * 2 * inner_prod(ctx.projected_state, mu).real
                    )

                if values[op.param_name].requires_grad:  # type: ignore [index]
                    grads_dict[op.param_name] = grad_out * 2 * -values[op.param_name]  # type: ignore [index]
                ctx.projected_state = apply_operator(
                    ctx.projected_state,
                    op.dagger(values, ctx.embedding),
                    op.qubit_support,
                )
            else:
                logger.error(
                    f"AdjointExpectation does not support operation: {type(op)}."
                )
        return (None, None, None, None, None, *grads_dict.values())



from __future__ import annotations

from abc import abstractmethod
from typing import Callable

import torch
from torch import Tensor

from pyqtorch.time_dependent.options import KrylovSolverOptions


class KrylovIntegrator:
    """Krylov subspace method to evolve the state vector."""

    def __init__(
        self,
        H: Callable,
        y0: Tensor,
        tsave: Tensor,
        options: KrylovSolverOptions,
    ):
        self.H = H
        self.t0 = 0.0
        self.y0 = y0
        self.tsave = tsave
        self.options = options

    @abstractmethod
    def integrate(self, t0: float, t1: float, y: Tensor) -> Tensor:
        pass

    def run(self) -> Tensor:

        out = []
        for i in range(self.y0.shape[1]):
            # run the Krylov routine
            result = []
            y = self.y0[:, i : i + 1]
            t = self.t0
            for tnext in self.tsave:
                y = self.integrate(t, tnext, y)
                result.append(y.T)
                t = tnext
            out.append(torch.cat(result))
        return torch.stack(out, dim=2)



from __future__ import annotations

import random

import pytest
import torch
from helpers import calc_mat_vec_wavefunction, get_op_support, random_pauli_hamiltonian

from pyqtorch.composite import Add, Scale, Sequence
from pyqtorch.hamiltonians import GeneratorType, HamiltonianEvolution
from pyqtorch.primitives import (
    CNOT,
    CPHASE,
    CRX,
    CRY,
    CRZ,
    CSWAP,
    CY,
    CZ,
    OPS_2Q,
    OPS_3Q,
    OPS_DIAGONAL,
    OPS_DIAGONAL_PARAM,
    OPS_DIGITAL,
    OPS_PARAM,
    OPS_PARAM_2Q,
    SWAP,
    N,
    Parametric,
    Primitive,
    Projector,
    Toffoli,
)
from pyqtorch.utils import (
    ATOL,
    RTOL,
    density_mat,
    permute_basis,
    random_state,
    todense_tensor,
)

pi = torch.tensor(torch.pi)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_primitive_tensor(
    n_qubits: int,
    make_param: bool,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits, k_1q, k_2q, make_param, diagonal=diagonal
    )
    values = {param: torch.rand(1) for param in param_list}
    full_support = tuple(range(n_qubits)) if use_full_support else None
    generator_matrix = generator.tensor(
        values=values, full_support=full_support, diagonal=diagonal
    )
    generator_matrix = (
        generator_matrix.repeat((1, 1, batch_size))
        if len(generator_matrix.size()) == 3
        else generator_matrix.repeat((1, batch_size))
    )
    diag_op = diagonal and generator.is_diagonal
    primitive_op = Primitive(
        generator_matrix,
        qubit_support=full_support if use_full_support else generator.qubit_support,  # type: ignore[arg-type]
        diagonal=diagonal,
    )
    if diag_op:
        assert len(generator_matrix.size()) == 2
        assert primitive_op.is_diagonal
    else:
        assert len(generator_matrix.size()) == 3
        assert not primitive_op.is_diagonal
    assert torch.allclose(primitive_op.tensor(), generator_matrix, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_dm", [True, False])
@pytest.mark.parametrize("use_permute", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_digital_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    use_permute: bool,
    use_dm: bool,
) -> None:
    """
    Goes through all non-parametric gates and tests their application to a random state
    in comparison with the `tensor` method, either using just the qubit support of the gate
    or expanding its matrix to the maximum qubit support of the full circuit.
    """
    op: type[Primitive]
    for op in OPS_DIGITAL:
        supp = get_op_support(op, n_qubits)
        op_concrete = op(*supp)
        if op in OPS_DIAGONAL:
            assert op_concrete.is_diagonal
        psi_init = random_state(n_qubits, batch_size)
        if use_dm:
            psi_star = op_concrete(density_mat(psi_init))
        else:
            psi_star = op_concrete(psi_init)
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op_concrete, psi_init, full_support=full_support, use_permute=use_permute
        )
        if use_dm:
            psi_expected = density_mat(psi_expected)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_dm", [True, False])
@pytest.mark.parametrize("use_permute", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("dtype", [torch.complex64, torch.complex128])
def test_param_tensor(
    n_qubits: int,
    batch_size: int,
    dtype: torch.dtype,
    use_full_support: bool,
    use_permute: bool,
    use_dm: bool,
) -> None:
    """
    Goes through all parametric gates and tests their application to a random state
    in comparison with the `tensor` method, either using just the qubit support of the gate
    or expanding its matrix to the maximum qubit support of the full circuit.
    """
    op: type[Parametric]
    for op in OPS_PARAM:
        supp = get_op_support(op, n_qubits)
        params = [f"th{i}" for i in range(op.n_params)]
        op_concrete = op(*supp, *params).to(dtype=dtype)
        if op in OPS_DIAGONAL_PARAM:
            assert op_concrete.is_diagonal
        psi_init = random_state(n_qubits, dtype=dtype)
        values = {param: torch.rand(batch_size) for param in params}
        if use_dm:
            psi_star = op_concrete(density_mat(psi_init), values)
        else:
            psi_star = op_concrete(psi_init, values)
        assert psi_star.dtype == dtype
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op_concrete,
            psi_init,
            values=values,
            full_support=full_support,
            use_permute=use_permute,
        )
        if use_dm:
            psi_expected = density_mat(psi_expected)
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("compose", [Sequence, Add])
def test_sequence_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    compose: type[Sequence] | type[Add],
) -> None:
    op_list = []
    values = {}
    op: type[Primitive] | type[Parametric]
    """
    Builds a Sequence or Add composition of all possible gates on random qubit
    supports. Also assigns a Scale of a random value to the non-parametric gates.
    Tests the forward method (which goes through each gate individually) to the
    `tensor` method, which builds the full operator matrix and applies it.
    """
    for op in OPS_DIGITAL:
        supp = get_op_support(op, n_qubits)
        op_concrete = Scale(op(*supp), torch.rand(1))
        if op in OPS_DIAGONAL:
            assert op_concrete.is_diagonal
        op_list.append(op_concrete)
    for op in OPS_PARAM:
        supp = get_op_support(op, n_qubits)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values.update({param: torch.rand(batch_size) for param in params})
        op_concrete = op(*supp, *params)
        if op in OPS_DIAGONAL_PARAM:
            assert op_concrete.is_diagonal
        op_list.append(op_concrete)
    random.shuffle(op_list)
    op_composite = compose(op_list)
    psi_init = random_state(n_qubits, batch_size)
    psi_star = op_composite(psi_init, values)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    psi_expected = calc_mat_vec_wavefunction(
        op_composite, psi_init, values=values, full_support=full_support
    )
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("compose", [Sequence, Add])
def test_diagonal_sequence_tensor(
    n_qubits: int,
    batch_size: int,
    use_full_support: bool,
    compose: type[Sequence] | type[Add],
) -> None:
    op_list = []
    values = {}
    op: type[Primitive] | type[Parametric]
    for op in OPS_DIAGONAL.intersection(OPS_DIGITAL):
        supp = get_op_support(op, n_qubits)
        op_concrete = Scale(op(*supp), torch.rand(1))
        op_list.append(op_concrete)
    for op in OPS_DIAGONAL_PARAM.intersection(OPS_PARAM):
        supp = get_op_support(op, n_qubits)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values.update({param: torch.rand(batch_size) for param in params})
        op_concrete = op(*supp, *params)
        op_list.append(op_concrete)
    random.shuffle(op_list)
    op_composite = compose(op_list)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    assert op_composite.is_diagonal
    tensor_composite = op_composite.tensor(values, full_support=full_support)
    tensor_composite_diagonal = op_composite.tensor(
        values, full_support=full_support, diagonal=True
    )
    dense_diagonal = todense_tensor(tensor_composite_diagonal)
    assert torch.allclose(tensor_composite, dense_diagonal, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("n_proj", [1, 3])
@pytest.mark.parametrize("batch_size", [1, 5])
def test_projector_tensor(
    n_qubits: int, n_proj: int, batch_size: int, use_full_support: bool
) -> None:
    """
    Instantiates various random projectors on arbitrary qubit support
    and compares the forward method with directly applying the tensor.
    """
    iterations = 5
    for _ in range(iterations):
        rand_int_1 = random.randint(0, 2**n_proj - 1)
        rand_int_2 = random.randint(0, 2**n_proj - 1)
        bitstring1 = "{0:b}".format(rand_int_1).zfill(n_proj)
        bitstring2 = "{0:b}".format(rand_int_2).zfill(n_proj)
        supp = tuple(random.sample(range(n_qubits), n_proj))
        op = Projector(supp, bitstring1, bitstring2)
        psi_init = random_state(n_qubits, batch_size)
        psi_star = op(psi_init)
        full_support = tuple(range(n_qubits)) if use_full_support else None
        psi_expected = calc_mat_vec_wavefunction(
            op, psi_init, full_support=full_support
        )
        assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("operator", [N, CNOT, SWAP])
@pytest.mark.parametrize("use_full_support", [True, False])
def test_projector_vs_operator(
    n_qubits: int,
    operator: type[Primitive],
    use_full_support: bool,
) -> None:
    if operator == N:
        supp: tuple = (random.randint(0, n_qubits - 1),)
        op_concrete = N(*supp)
        projector = Projector(supp, "1", "1")
    if operator == CNOT:
        supp = tuple(random.sample(range(n_qubits), 2))
        op_concrete = CNOT(*supp)
        projector = Add(
            [
                Projector(supp, "00", "00"),
                Projector(supp, "01", "01"),
                Projector(supp, "10", "11"),
                Projector(supp, "11", "10"),
            ]
        )
    if operator == SWAP:
        supp = tuple(random.sample(range(n_qubits), 2))
        op_concrete = SWAP(*supp)
        projector = Add(
            [
                Projector(supp, "00", "00"),
                Projector(supp, "01", "10"),
                Projector(supp, "10", "01"),
                Projector(supp, "11", "11"),
            ]
        )
    full_support = tuple(range(n_qubits)) if use_full_support else None
    projector_mat = projector.tensor(full_support=full_support)
    operator_mat = op_concrete.tensor(full_support=full_support)
    assert torch.allclose(projector_mat, operator_mat, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("make_param", [True, False])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_hevo_pauli_tensor(
    n_qubits: int,
    make_param: bool,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * n_qubits  # Number of 1-qubit terms
    k_2q = n_qubits**2  # Number of 2-qubit terms
    generator, param_list = random_pauli_hamiltonian(
        n_qubits, k_1q, k_2q, make_param, diagonal=diagonal
    )

    values = {param: torch.rand(batch_size) for param in param_list}
    psi_init = random_state(n_qubits, batch_size)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    # Test the generator itself
    psi_star = generator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(generator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)

    # Test the hamiltonian evolution
    tparam = "t"
    operator = HamiltonianEvolution(generator, tparam)
    if diagonal:
        assert generator.is_diagonal
        if not make_param:
            assert operator.is_diagonal

    if make_param:
        assert operator.generator_type in (
            GeneratorType.PARAMETRIC_OPERATION,
            GeneratorType.PARAMETRIC_COMMUTING_SEQUENCE,
        )
    else:
        assert operator.generator_type in (
            GeneratorType.OPERATION,
            GeneratorType.COMMUTING_SEQUENCE,
        )
    values[tparam] = torch.rand(batch_size)
    psi_star = operator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(operator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("gen_qubits", [3, 4])
@pytest.mark.parametrize("n_qubits", [4, 5])
@pytest.mark.parametrize("use_full_support", [True, False])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_hevo_tensor_tensor(
    gen_qubits: int,
    n_qubits: int,
    use_full_support: bool,
    batch_size: int,
    diagonal: bool,
) -> None:
    k_1q = 2 * gen_qubits  # Number of 1-qubit terms
    k_2q = gen_qubits**2  # Number of 2-qubit terms
    generator, _ = random_pauli_hamiltonian(gen_qubits, k_1q, k_2q, diagonal=diagonal)
    psi_init = random_state(n_qubits, batch_size)
    full_support = tuple(range(n_qubits)) if use_full_support else None
    # Test the hamiltonian evolution
    generator_matrix = generator.tensor()
    supp = tuple(random.sample(range(n_qubits), gen_qubits))
    tparam = "t"
    operator = HamiltonianEvolution(generator_matrix, tparam, supp)
    assert operator.generator_type == GeneratorType.TENSOR
    if diagonal:
        assert generator.is_diagonal
        assert operator.is_diagonal
    values = {tparam: torch.rand(batch_size)}
    psi_star = operator(psi_init, values)
    psi_expected = calc_mat_vec_wavefunction(operator, psi_init, values, full_support)
    assert torch.allclose(psi_star, psi_expected, rtol=RTOL, atol=ATOL)


@pytest.mark.parametrize("n_qubits", [3, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_permute_tensor(n_qubits: int, diagonal: bool) -> None:
    ops = OPS_2Q.union(OPS_3Q) if not diagonal else {CZ}
    for op in ops:
        supp, ordered_supp = get_op_support(op, n_qubits, get_ordered=True)

        op_concrete1 = op(*supp)
        op_concrete2 = op(*ordered_supp)

        mat1 = op_concrete1.tensor(diagonal=diagonal)
        mat2 = op_concrete2.tensor(diagonal=diagonal)

        perm = op_concrete1._qubit_support.qubits

        assert torch.allclose(
            mat1, permute_basis(mat2, perm, inv=True, diagonal=diagonal)
        )
        assert torch.allclose(mat2, permute_basis(mat1, perm, diagonal=diagonal))


@pytest.mark.parametrize("n_qubits", [3, 5])
@pytest.mark.parametrize("batch_size", [1, 5])
@pytest.mark.parametrize("diagonal", [False, True])
def test_permute_tensor_parametric(
    n_qubits: int, batch_size: int, diagonal: bool
) -> None:
    ops = (
        OPS_PARAM_2Q if not diagonal else OPS_PARAM_2Q.intersection(OPS_DIAGONAL_PARAM)
    )
    for op in ops:
        supp, ordered_supp = get_op_support(op, n_qubits, get_ordered=True)
        params = [f"{op.__name__}_th{i}" for i in range(op.n_params)]
        values = {param: torch.rand(batch_size) for param in params}

        op_concrete1 = op(*supp, *params)
        op_concrete2 = op(*ordered_supp, *params)

        mat1 = op_concrete1.tensor(values=values, diagonal=diagonal)
        mat2 = op_concrete2.tensor(values=values, diagonal=diagonal)

        perm = op_concrete1._qubit_support.qubits

        assert torch.allclose(
            mat1, permute_basis(mat2, perm, inv=True, diagonal=diagonal)
        )
        assert torch.allclose(mat2, permute_basis(mat1, perm, diagonal=diagonal))


def test_tensor_symmetries() -> None:
    assert not torch.allclose(CNOT(0, 1).tensor(), CNOT(1, 0).tensor())
    assert not torch.allclose(CY(0, 1).tensor(), CY(1, 0).tensor())
    assert not torch.allclose(CZ(0, 1).tensor(), CY(1, 0).tensor())
    assert not torch.allclose(CRX(0, 1, 1.0).tensor(), CRX(1, 0, 1.0).tensor())
    assert not torch.allclose(CRY(0, 1, 1.0).tensor(), CRY(1, 0, 1.0).tensor())
    assert not torch.allclose(CRZ(0, 1, 1.0).tensor(), CRZ(1, 0, 1.0).tensor())
    assert torch.allclose(CPHASE(0, 1, 1.0).tensor(), CPHASE(1, 0, 1.0).tensor())
    assert torch.allclose(SWAP(0, 1).tensor(), SWAP(1, 0).tensor())
    assert torch.allclose(CSWAP(0, (1, 2)).tensor(), CSWAP(0, (2, 1)).tensor())
    assert torch.allclose(CSWAP(1, (0, 2)).tensor(), CSWAP(1, (2, 0)).tensor())
    assert torch.allclose(CSWAP(2, (0, 1)).tensor(), CSWAP(2, (1, 0)).tensor())
    assert torch.allclose(Toffoli((0, 1), 2).tensor(), Toffoli((1, 0), 2).tensor())
    assert torch.allclose(Toffoli((0, 2), 1).tensor(), Toffoli((2, 0), 1).tensor())
    assert torch.allclose(Toffoli((1, 2), 0).tensor(), Toffoli((2, 1), 0).tensor())



from __future__ import annotations

from typing import Any, Callable

from torch import Tensor

from pyqtorch.time_dependent.options import AdaptiveSolverOptions, KrylovSolverOptions
from pyqtorch.time_dependent.solvers import SEDormandPrince5, SEKrylov
from pyqtorch.utils import Result, SolverType


def sesolve(
    H: Callable[[float], Tensor],
    psi0: Tensor,
    tsave: Tensor,
    solver: SolverType,
    options: dict[str, Any] | None = None,
) -> Result:
    """Solve time-dependent Schrodinger equation.

    Args:
        H (Callable[[float], Tensor]): time-dependent Hamiltonian of the system
        psi0 (Tensor): initial state of the system
        tsave (Tensor): tensor containing simulation time instants
        solver (SolverType): name of the solver to use
        options (dict[str, Any], optional): additional options passed to the solver.
            Defaults to None.

    Returns:
        Result: dataclass containing the simulated states at each time moment
    """
    # check dimensions of initial state
    n = H(0.0).shape[0]
    if (psi0.shape[0] != n) or len(psi0.shape) != 2:
        raise ValueError(
            f"Argument `psi0` must be a 2D tensor of shape `({n}, batch_size)`. Current shape:"
            f" {tuple(psi0.shape)}."
        )

    options = options or dict()
    # instantiate appropriate solver
    if solver == SolverType.DP5_SE:
        opt = AdaptiveSolverOptions(**options)
        s = SEDormandPrince5(H, psi0, tsave, opt)
    elif solver == SolverType.KRYLOV_SE:
        opt = KrylovSolverOptions(**options)  # type: ignore [assignment]
        s = SEKrylov(H, psi0, tsave, opt)  # type: ignore [assignment, arg-type]
    else:
        raise ValueError("Requested solver is not available.")

    # compute the result
    result = s.run()

    return Result(result)



from __future__ import annotations

from importlib import import_module
from logging import getLogger
from typing import Any, Tuple

from numpy.typing import ArrayLike, DTypeLike
from torch import Tensor

logger = getLogger(__name__)


ARRAYLIKE_FN_MAP = {
    "torch": ("torch", "tensor"),
    "jax": ("jax.numpy", "array"),
    "numpy": ("numpy", "array"),
}


DEFAULT_JAX_MAPPING = {
    "mul": ("jax.numpy", "multiply"),
    "sub": ("jax.numpy", "subtract"),
    "div": ("jax.numpy", "divide"),
}
DEFAULT_TORCH_MAPPING = {"hs": ("pyqtorch.utils", "heaviside")}
DEFAULT_NUMPY_MAPPING = {
    "mul": ("numpy", "multiply"),
    "sub": ("numpy", "subtract"),
    "div": ("numpy", "divide"),
}

DEFAULT_INSTRUCTION_MAPPING = {
    "torch": DEFAULT_TORCH_MAPPING,
    "jax": DEFAULT_JAX_MAPPING,
    "numpy": DEFAULT_NUMPY_MAPPING,
}


class ConcretizedCallable:
    """Transform an abstract function name and arguments into
        a callable in a linear algebra engine which can be evaluated
        using user input.
    Arguments:
        call_name: The name of the function.
        abstract_args: A list of arguments to the function,
                       can be numeric types for constants or strings for parameters
        instruction_mapping: A dict containing user-passed mappings from a function name
                            to its implementation.
        engine_name: The name of the framework to use.
        device: Which device to use.

    Example:
    ```
    import torch

    from pyqtorch.embed import ConcretizedCallable


    In [11]: call = ConcretizedCallable('sin', ['x'], engine_name='numpy')
    In [12]: call({'x': 0.5})
    Out[12]: 0.479425538604203

    In [13]: call = ConcretizedCallable('sin', ['x'], engine_name='torch')
    In [14]: call({'x': torch.rand(1)})
    Out[14]: tensor([0.5531])

    In [15]: call = ConcretizedCallable('sin', ['x'], engine_name='jax')
    In [16]: call({'x': 0.5})
    Out[16]: Array(0.47942555, dtype=float32, weak_type=True)
    ```



    """

    def __init__(
        self,
        call_name: str = "",
        abstract_args: list[str | float | int | complex | ConcretizedCallable] = ["x"],
        instruction_mapping: dict[str, Tuple[str, str]] | None = None,
        engine_name: str = "torch",
        device: str = "cpu",
        dtype: Any = None,
    ) -> None:
        instruction_mapping = instruction_mapping or dict()
        instruction_mapping = {
            **instruction_mapping,
            **DEFAULT_INSTRUCTION_MAPPING[engine_name],
        }
        self.call_name = call_name
        self.abstract_args = abstract_args
        self.engine_name = engine_name
        self._device = device
        self._dtype = dtype
        self.engine_call = None
        engine = None
        if not all(
            [
                isinstance(arg, (str, float, int, complex, Tensor, ConcretizedCallable))
                for arg in abstract_args
            ]
        ):
            raise TypeError(
                "Only str, float, int, complex, Tensor or ConcretizedCallable type elements \
                are supported for abstract_args"
            )
        try:
            engine_name, fn_name = ARRAYLIKE_FN_MAP[engine_name]
            engine = import_module(engine_name)
            self.arraylike_fn = getattr(engine, fn_name)
        except (ModuleNotFoundError, ImportError) as e:
            logger.error(f"Unable to import {engine_name} due to {e}.")

        try:
            self.engine_call = getattr(engine, call_name, None)
            if self.engine_call is None:
                mod, fn = instruction_mapping[call_name]
                self.engine_call = getattr(import_module(mod), fn)
        except (ImportError, KeyError) as e:
            logger.error(
                f"Requested function {call_name} can not be imported from {engine_name} and is"
                + f" not in instruction_mapping {instruction_mapping} due to {e}."
            )

    def evaluate(self, inputs: dict[str, ArrayLike] | None = None) -> ArrayLike:
        arraylike_args = []
        inputs = inputs or dict()
        for symbol_or_numeric in self.abstract_args:
            if isinstance(symbol_or_numeric, ConcretizedCallable):
                arraylike_args.append(symbol_or_numeric(inputs))
            if isinstance(symbol_or_numeric, (float, int, Tensor)):
                arraylike_args.append(
                    self.arraylike_fn(symbol_or_numeric, device=self.device)
                )
            elif isinstance(symbol_or_numeric, str):
                arraylike_args.append(inputs[symbol_or_numeric])
        return self.engine_call(*arraylike_args)  # type: ignore[misc]

    @classmethod
    def _get_independent_args(cls, cc: ConcretizedCallable) -> set:
        out: set = set()
        if len(cc.abstract_args) == 1 and isinstance(cc.abstract_args[0], str):
            return set([cc.abstract_args[0]])
        else:
            for arg in cc.abstract_args:
                if isinstance(arg, ConcretizedCallable):
                    res = cls._get_independent_args(arg)
                    out = out.union(res)
                else:
                    if isinstance(arg, str):
                        out.add(arg)
        return out

    @property
    def independent_args(self) -> list:
        return list(self._get_independent_args(self))

    def __call__(self, inputs: dict[str, ArrayLike] | None = None) -> ArrayLike:
        return self.evaluate(inputs)

    def __mul__(
        self, other: str | int | float | complex | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("mul", [self, other])

    def __rmul__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("mul", [other, self])

    def __add__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("add", [self, other])

    def __radd__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("add", [other, self])

    def __sub__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("sub", [self, other])

    def __rsub__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("sub", [other, self])

    def __pow__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("pow", [self, other])

    def __rpow__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("pow", [other, self])

    def __truediv__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("div", [self, other])

    def __rtruediv__(
        self, other: str | int | float | ConcretizedCallable
    ) -> ConcretizedCallable:
        return ConcretizedCallable("div", [other, self])

    def __repr__(self) -> str:
        return f"{self.call_name}({self.abstract_args})"

    def __neg__(self) -> ConcretizedCallable:
        return -1 * self

    @property
    def device(self) -> str:
        return self._device

    @property
    def dtype(self) -> Any:
        return self._dtype

    def to(self, *args: Any, **kwargs: Any) -> ConcretizedCallable:
        self._device = kwargs.get("device", None)
        self._dtype = kwargs.get("dtype", None)
        return self


def init_param(
    engine_name: str, trainable: bool = True, device: str = "cpu"
) -> ArrayLike:
    engine = import_module(engine_name)
    if engine_name == "jax":
        return engine.random.uniform(engine.random.PRNGKey(42), shape=(1,))
    elif engine_name == "torch":
        return engine.rand(1, requires_grad=trainable, device=device)
    elif engine_name == "numpy":
        return engine.random.uniform(0, 1)


def sin(x: str | ConcretizedCallable):
    return ConcretizedCallable("sin", [x])


def cos(x: str | ConcretizedCallable):
    return ConcretizedCallable("cos", [x])


def log(x: str | ConcretizedCallable):
    return ConcretizedCallable("log", [x])


def tan(x: str | ConcretizedCallable):
    return ConcretizedCallable("tan", [x])


def tanh(x: str | ConcretizedCallable):
    return ConcretizedCallable("tanh", [x])


def sqrt(x: str | ConcretizedCallable):
    return ConcretizedCallable("sqrt", [x])


class Embedding:
    """A class relating variational and feature parameters used in ConcretizedCallable instances to
    parameter names used in gates.

    Arguments:
        vparam_names: A list of variational parameters.
        fparam_names: A list of feature parameters.
        var_to_call: A dict mapping from <`parameter_name`: ConcretizedCallable> pairs,.
        tparam_name: Optional name for a time parameter.
        engine_name: The name of the linear algebra engine.
        device: The device to use

    Example:
    ```
    from __future__ import annotations

    import numpy as np
    import pytest
    import torch
    import torch.autograd.gradcheck

    import pyqtorch as pyq
    from pyqtorch.embed import ConcretizedCallable, Embedding
    name0, fn0 = "fn0", ConcretizedCallable("sin", ["x"])
    name1, fn1 = "fn1", ConcretizedCallable("mul", ["fn0", "y"])
    name2, fn2 = "fn2", ConcretizedCallable("mul", ["fn1", 2.0])
    name3, fn3 = "fn3", ConcretizedCallable("log", ["fn2"])
    embedding = pyq.Embedding(
        vparam_names=["x"],
        fparam_names=["y"],
        var_to_call={name0: fn0, name1: fn1, name2: fn2, name3: fn3},
    )
    rx = pyq.RX(0, param_name=name0)
    cry = pyq.CRY(0, 1, param_name=name1)
    phase = pyq.PHASE(1, param_name=name2)
    ry = pyq.RY(1, param_name=name3)
    cnot = pyq.CNOT(1, 2)
    ops = [rx, cry, phase, ry, cnot]
    n_qubits = 3
    circ = pyq.QuantumCircuit(n_qubits, ops)
    obs = pyq.Observable([pyq.Z(0)])

    state = pyq.zero_state(n_qubits)

    x = torch.rand(1, requires_grad=True)
    y = torch.rand(1, requires_grad=True)

    values_ad = {"x": x, "y": y}
    embedded_params = embedding(values_ad)
    wf = pyq.run(circ, state, embedded_params, embedding)
    ```
    """

    def __init__(
        self,
        vparam_names: list[str] = [],
        fparam_names: list[str] = [],
        var_to_call: dict[str, ConcretizedCallable] | None = None,
        tparam_name: str | None = None,
        engine_name: str = "torch",
        device: str = "cpu",
    ) -> None:
        var_to_call = var_to_call or dict()
        self.vparams = {
            vp: init_param(engine_name, trainable=True, device=device)
            for vp in vparam_names
        }
        self.fparam_names: list[str] = fparam_names
        self.tparam_name = tparam_name
        self.var_to_call: dict[str, ConcretizedCallable] = var_to_call
        self._dtype: DTypeLike = None
        self.tracked_vars: list[str] = []
        self._device = device
        self._tracked_vars_identified = False
        self.engine_name = engine_name

    @property
    def root_param_names(self) -> list[str]:
        return list(self.vparams.keys()) + self.fparam_names

    def embed_all(
        self,
        inputs: dict[str, ArrayLike] | None = None,
    ) -> dict[str, ArrayLike]:
        """The standard embedding of all intermediate and leaf parameters.
        Include the root_params, i.e., the vparams and fparams original values
        to be reused in computations.
        """
        inputs = inputs or dict()
        for intermediate_or_leaf_var, engine_callable in self.var_to_call.items():
            # We mutate the original inputs dict and include intermediates and leaves.
            if not self._tracked_vars_identified:
                # we do this only on the first embedding call
                if self.tparam_name and any(
                    [
                        p in [self.tparam_name] + self.tracked_vars
                        for p in engine_callable.abstract_args
                    ]  # we check if any parameter in the callables args is time
                    # or depends on an intermediate variable which itself depends on time
                ):
                    self.tracked_vars.append(intermediate_or_leaf_var)
                    # we remember which parameters depend on time
            inputs[intermediate_or_leaf_var] = engine_callable(inputs)
        self._tracked_vars_identified = True
        return inputs

    def reembed_tparam(
        self,
        embedded_params: dict[str, ArrayLike],
        tparam_value: ArrayLike,
    ) -> dict[str, ArrayLike]:
        """Receive already embedded params containing intermediate and leaf parameters
        and recalculate the those which are dependent on `tparam_name` using the new value
        `tparam_value`.
        """
        if self.tparam_name is None:
            raise ValueError(
                "`reembed_param` requires a `tparam_name` to be passed\
                              when initializing the `Embedding` class"
            )
        embedded_params[self.tparam_name] = tparam_value
        for time_dependent_param in self.tracked_vars:
            embedded_params[time_dependent_param] = self.var_to_call[
                time_dependent_param
            ](embedded_params)
        return embedded_params

    def __call__(
        self, inputs: dict[str, ArrayLike] | None = None
    ) -> dict[str, ArrayLike]:
        """Functional version of legacy embedding: Return a new dictionary\
        with all embedded parameters."""
        return self.embed_all(inputs or dict())

    @property
    def dtype(self) -> DTypeLike:
        return self._dtype

    @property
    def device(self) -> str:
        return self._device

    def to(self, *args: Any, **kwargs: Any) -> Embedding:
        if self.engine_name == "torch":
            # we only support this for torch for now
            self.vparams = {p: t.to(*args, **kwargs) for p, t in self.vparams.items()}
            self.var_to_call = {
                p: call.to(*args, **kwargs) for p, call in self.var_to_call.items()
            }
            # Dtype and device have to be passes as kwargs
            self._dtype = kwargs.get("dtype", self._dtype)
            self._device = kwargs.get("device", self._device)
        return self
